#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import struct

import netius.common

class TFTPSession(object):

    def __init__(self, owner, name = None, mode = None):
        self.owner = owner
        self.name = name
        self.mode = mode
        self.file = None
        self.completed = False
        self.sequence = 0

    def close(self):
        self.reset()

    def reset(self):
        if self.file: self.file.close()
        self.name = None
        self.mode = None
        self.file = None
        self.completed = False
        self.sequence = 0

    def next(self, size = 512, increment = True):
        if self.completed: return None
        file = self._get_file()
        data = file.read(size)
        self.completed = len(data) < size
        if increment: self.increment()
        header = struct.pack("!HH", netius.common.DATA_TFTP, self.sequence)
        return header + data

    def ack(self, size = 512, increment = True):
        if self.sequence == 0: return None
        return self.next(size = size, increment = increment)

    def increment(self):
        self.sequence += 1

    def get_info(self):
        buffer = netius.legacy.StringIO()
        buffer.write("name      := %s\n" % self.name)
        buffer.write("mode      := %s\n" % self.mode)
        buffer.write("completed := %d\n" % self.completed)
        buffer.write("sequence  := %d" % self.sequence)
        buffer.seek(0)
        info = buffer.read()
        return info

    def print_info(self):
        info = self.get_info()
        print(info)

    def _get_file(self, allow_absolute = False):
        if self.file: return self.file
        if not allow_absolute: name = self.name.lstrip("/")
        path = os.path.join(self.owner.base_path, name)
        self.file = open(path, "rb")
        return self.file

class TFTPRequest(object):

    parsers_m = None
    parsers_l = None

    def __init__(self, data, session):
        self.data = data
        self.session = session

        cls = self.__class__
        cls.generate()

    @classmethod
    def generate(cls):
        if cls.parsers_m: return
        cls.parsers_m = (
            cls._parse_rrq,
            cls._parse_wrq,
            cls._parse_data,
            cls._parse_ack,
            cls._parse_error
        )
        cls.parsers_l = len(cls.parsers_m)

    def get_info(self):
        session_info = self.session.get_info()
        buffer = netius.legacy.StringIO()
        buffer.write("op        := %d\n" % self.op)
        buffer.write("payload   := %s" % repr(self.payload))
        if session_info: buffer.write("\n" + session_info)
        buffer.seek(0)
        info = buffer.read()
        return info

    def print_info(self):
        info = self.get_info()
        print(info)

    def parse(self):
        cls = self.__class__

        format = "!H"

        self.header = self.data[:2]
        self.payload = self.data[2:]
        result = struct.unpack(format, self.header)
        self.op = result[0]

        method = cls.parsers_m[self.op - 1]
        method(self)

    def get_type(self):
        return self.op

    def get_type_s(self):
        type = self.get_type()
        type_s = netius.common.TYPES_TFTP.get(type, None)
        return type_s

    def response(self, options = {}):
        if self.op == netius.common.ACK_TFTP: return self.session.ack()
        return self.session.next()

    @classmethod
    def _parse_rrq(cls, self):
        payload = self.payload
        self.session.reset()
        self.session.name, payload = cls._str(payload)
        self.session.mode, payload = cls._str(payload)

    @classmethod
    def _parse_wrq(cls, self):
        raise netius.NotImplemented("Operation not implemented")

    @classmethod
    def _parse_data(cls, self):
        raise netius.NotImplemented("Operation not implemented")

    @classmethod
    def _parse_ack(cls, self):
        pass

    @classmethod
    def _parse_error(cls, self):
        raise netius.NotImplemented("Operation not implemented")

    @classmethod
    def _str(cls, data):
        index = data.index(b"\x00")
        value, remaining = data[:index], data[index + 1:]
        value = netius.legacy.str(value)
        return value, remaining

class TFTPServer(netius.DatagramServer):
    """
    Abstract trivial ftp server implementation that handles simple
    file system based file serving.

    :see: http://tools.ietf.org/html/rfc1350
    """

    ALLOWED_OPERATIONS = (
        netius.common.RRQ_TFTP,
        netius.common.ACK_TFTP
    )

    def __init__(self, base_path = "", *args, **kwargs):
        netius.DatagramServer.__init__(self, *args, **kwargs)
        self.base_path = base_path
        self.sessions = dict()

    def serve(self, port = 69, *args, **kwargs):
        netius.DatagramServer.serve(self, port = port, *args, **kwargs)

    def on_data(self, address, data):
        netius.DatagramServer.on_data(self, address, data)

        try:
            session = self.sessions.get(address, None)
            if not session: session = TFTPSession(self)
            self.sessions[address] = session

            request = TFTPRequest(data, session)
            request.parse()

            self.on_data_tftp(address, request)
        except Exception as exception:
            self.on_error_tftp(address, exception)

    def on_serve(self):
        netius.DatagramServer.on_serve(self)
        if self.env: self.base_path = self.get_env("BASE_PATH", self.base_path)
        self.info("Starting TFTP server ...")
        self.info("Defining '%s' as the root of the file server ..." % (self.base_path or "."))

    def on_data_tftp(self, address, request):
        cls = self.__class__

        type = request.get_type()
        type_s = request.get_type_s()

        self.debug("Received %s message from '%s'" % (type_s, address))

        if not type in cls.ALLOWED_OPERATIONS:
            raise netius.NetiusError(
                "Invalid operation type '%d'", type
            )

        response = request.response()
        if not response: return

        self.send(response, address)

    def on_error_tftp(self, address, exception):
        message = str(exception)
        message_b = netius.legacy.bytes(message)
        header = struct.pack("!HH", netius.common.ERROR_TFTP, 0)
        response = header + message_b + b"\x00"
        self.send(response, address)
        self.info("Sent error message '%s' to '%s'" % (message, address))

if __name__ == "__main__":
    import logging
    server = TFTPServer(level = logging.DEBUG)
    server.serve(env = True)
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import heapq

class PriorityDict(dict):

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self._rebuild_heap()

    def __setitem__(self, key, val):
        dict.__setitem__(self, key, val)

        if len(self._heap) < len(self) * 2:
            heapq.heappush(self._heap, (val, key))
        else:
            self._rebuild_heap()

    def smallest(self):
        heap = self._heap
        v, k = heap[0]
        while not k in self or self[k] != v:
            heapq.heappop(heap)
            v, k = heap[0]
        return k

    def pop_smallest(self):
        heap = self._heap
        v, k = heapq.heappop(heap)
        while not k in self or self[k] != v:
            v, k = heapq.heappop(heap)
        del self[k]
        return k

    def setdefault(self, key, val):
        if not key in self:
            self[key] = val
            return val
        return self[key]

    def update(self, *args, **kwargs):
        dict.update(self, *args, **kwargs)
        self._rebuild_heap()

    def sorted_iter(self):
        while self: yield self.pop_smallest()

    def _rebuild_heap(self):
        self._heap = [(v, k) for k, v in self.items()]
        heapq.heapify(self._heap)

def file_iterator(file_object, chunk_size = 40960):
    file_object.seek(0, os.SEEK_END)
    size = file_object.tell()
    file_object.seek(0, os.SEEK_SET)
    yield size
    while True:
        data = file_object.read(chunk_size)
        if not data: break
        yield data



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import netius.clients

def http_static():
    request = netius.clients.HTTPClient.get_s(
        "https://www.flickr.com/",
        asynchronous = False
    )
    print(request["data"])

def http_callback():
    def callback(connection, parser, request):
        print(request["data"])
        client = connection.owner
        client.close()

    client = netius.clients.HTTPClient()
    client.get("https://www.flickr.com/", on_result = callback)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from . import asn
from . import calc
from . import dhcp
from . import dkim
from . import ftp
from . import geo
from . import http
from . import http2
from . import mime
from . import parser
from . import pop
from . import rsa
from . import setup
from . import smtp
from . import socks
from . import tls
from . import stream
from . import structures
from . import style
from . import tftp
from . import torrent
from . import util
from . import ws

from .asn import asn1_parse, asn1_length, asn1_gen, asn1_build
from .calc import prime, is_prime, relatively_prime, gcd, egcd, modinv,\
    random_integer_interval, random_primality, jacobi_witness, jacobi, ceil_integer
from .dhcp import SUBNET_DHCP, ROUTER_DHCP, DNS_DHCP, NAME_DHCP, BROADCAST_DHCP,\
    REQUESTED_DHCP, LEASE_DHCP, DISCOVER_DHCP, OFFER_DHCP, REQUEST_DHCP, DECLINE_DHCP,\
    ACK_DHCP, NAK_DHCP, IDENTIFIER_DHCP, RENEWAL_DHCP, REBIND_DHCP, PROXY_DHCP,\
    END_DHCP, OPTIONS_DHCP, TYPES_DHCP, VERBS_DHCP, AddressPool
from .dkim import dkim_sign, dkim_headers, dkim_body, dkim_fold, dkim_generate
from .ftp import FTPParser
from .geo import GeoResolver
from .http import REQUEST, RESPONSE, PLAIN_ENCODING, CHUNKED_ENCODING, GZIP_ENCODING,\
    DEFLATE_ENCODING, HTTP_09, HTTP_10, HTTP_11, VERSIONS_MAP, CODE_STRINGS, HTTPParser,\
    HTTPResponse
from .http2 import DATA, HEADERS, PRIORITY, RST_STREAM, SETTINGS, PUSH_PROMISE,\
    PING, GOAWAY, WINDOW_UPDATE, CONTINUATION, HTTP2_WINDOW, HTTP2_PREFACE,\
    HTTP2_TUPLES, HTTP2_NAMES, HTTP2_SETTINGS, HTTP2_SETTINGS_OPTIMAL, HTTP2_SETTINGS_T,\
    HTTP2_SETTINGS_OPTIMAL_T, HTTP2Parser, HTTP2Stream
from .mime import rfc822_parse, rfc822_join, mime_register
from .parser import Parser
from .pop import POPParser
from .rsa import open_pem_key, open_pem_data, write_pem_key, open_private_key, open_private_key_b64,\
    open_private_key_data, open_public_key, open_public_key_b64, open_public_key_data,\
    write_private_key, write_public_key, asn_private_key, asn_public_key, pem_to_der,\
    pem_limiters, private_to_public, assert_private, rsa_private, rsa_primes, rsa_exponents,\
    rsa_bits, rsa_sign, rsa_verify, rsa_crypt_s, rsa_crypt
from .setup import ensure_setup, ensure_ca
from .smtp import SMTPParser
from .socks import SOCKSParser
from .tls import TLSContextDict, LetsEncryptDict
from .stream import Stream, FileStream, FilesStream
from .structures import PriorityDict, file_iterator
from .style import BASE_STYLE
from .tftp import RRQ_TFTP, WRQ_TFTP, DATA_TFTP, ACK_TFTP, ERROR_TFTP, TYPES_TFTP
from .torrent import info_hash, bencode, bdecode, chunk, dechunk, TorrentParser
from .util import cstring, chunks, header_down, header_up, is_ip4, is_ip6, assert_ip4,\
    in_subnet_ip4, addr_to_ip4, addr_to_ip6, ip4_to_addr, string_to_bits, integer_to_bytes,\
    bytes_to_integer, random_integer, host, hostname, size_round_unit, verify, verify_equal,\
    verify_not_equal, verify_many
from .ws import encode_ws, decode_ws, assert_ws



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os

import netius

from . import http

class MJPGProtocol(http.HTTPProtocol):

    MAGIC_JPEG = b"\xff\xd8\xff\xe0"
    """ The magic signature for the jpeg infra-structure, this
    sequence of bytes is going to be used to detect new frames
    coming from the http based stream """

    EOI_JPEG = b"\xff\xd9"
    """ The sequence of bytes that indicate the end of the current
    image, when these bytes are detected on the stream the message
    should be "flushed" to the current output (emit) """

    def __init__(self, *args, **kwargs):
        http.HTTPProtocol.__init__(self, *args, **kwargs)
        self.buffer_l = []

    def add_buffer(self, data):
        self.buffer_l.append(data)

    def get_buffer(self, delete = True):
        if not self.buffer_l: return b""
        buffer = b"".join(self.buffer_l)
        if delete: del self.buffer_l[:]
        return buffer

    def on_partial(self, data):
        http.HTTPProtocol.on_partial(self, data)

        # retrieves the reference to the top class that is going to
        # be used for the correct parsing of the image
        cls = self.__class__

        # tries to find the end of image (EOI) indicator in the current
        # received data, and in case it's not found add the (partial)
        # data to the current buffer, to be latter processed
        eoi_index = data.find(cls.EOI_JPEG)
        if eoi_index == -1: self.buffer_l.append(data); return

        # calculates the size of the end of image (EOI) token so that
        # this value will be used for the calculus of the image data
        eoi_size = len(cls.EOI_JPEG)

        # adds the partial valid data of the current chunk to the buffer
        # and then joins the current buffer as the frame data, removing
        # the multipart header from it (to become a valid image)
        self.buffer_l.append(data[:eoi_index + eoi_size])
        frame = b"".join(self.buffer_l)
        multipart_index = frame.find(b"\r\n\r\n")
        frame = frame[multipart_index + 4:]

        # clears the current buffer and adds the remaining part of the
        # current chunk, that may be already part of a new image
        del self.buffer_l[:]
        self.buffer_l.append(data[eoi_index + eoi_size:])

        # calls the proper event handler for the new frame data that has
        # just been received, triggering the processing of the frame
        self.on_frame_mjpg(frame)

    def on_frame_mjpg(self, data):
        self.trigger("frame", self, data)

class MJPGClient(http.HTTPClient):

    protocol = MJPGProtocol

if __name__ == "__main__":
    index = 0
    limit = 30

    def on_frame(protocol, data):
        global index
        index += 1
        if index >= limit: return protocol.close()
        base_path = netius.conf("IMAGES_PATH", "images")
        base_path = os.path.abspath(base_path)
        base_path = os.path.normpath(base_path)
        if not os.path.exists(base_path): os.makedirs(base_path)
        path = os.path.join(base_path, "%08d.jpg" % index)
        file = open(path, "wb")
        try: file.write(data)
        finally: file.close()
        print("Saved frame %08d of %d bytes" % (index, len(data)))

    def on_finish(protocol):
        netius.compat_loop(loop).stop()

    url = netius.conf("MJPG_URL", "http://euglena.stanford.edu:20005/?action=stream")

    client = MJPGClient()
    loop, protocol = client.get(url)

    protocol.bind("frame", on_frame)
    protocol.bind("finish", on_finish)

    loop.run_forever()
    loop.close()
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import netius

class RawProtocol(netius.StreamProtocol):

    def send_basic(self):
        """
        Sends a basic HTTP 1.0 request, that can be used to
        run a simple operation on the raw protocol.
        """

        self.send("GET / HTTP/1.0\r\n\r\n")

class RawClient(netius.ClientAgent):

    protocol = RawProtocol

    @classmethod
    def run_s(
        cls,
        host,
        port = 8080,
        loop = None,
        *args,
        **kwargs
    ):
        protocol = cls.protocol()

        loop = netius.connect_stream(
            lambda: protocol,
            host = host,
            port = port,
            loop = loop,
            *args,
            **kwargs
        )

        return loop, protocol

if __name__ == "__main__":

    def on_open(protocol):
        protocol.send_basic()

    def on_data(protocol, data):
        print(data)

    def on_finsh(protocol):
        netius.compat_loop(loop).stop()

    loop, protocol = RawClient.run_s("localhost")

    protocol.bind("open", on_open)
    protocol.bind("data", on_data)
    protocol.bind("finish", on_finsh)

    loop.run_forever()
    loop.close()

else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import struct

import netius.common

DNS_QUERY = 0x0

DNS_RESPONSE = 0x1

DNS_SQUERY = 0x0

DNS_IQUERY = 0x1

DNS_STATUS = 0x2

DNS_AA = 0x04

DNS_TC = 0x02

DNS_RD = 0x01

DNS_TYPES = dict(
    A = 0x01,
    NS = 0x02,
    MD = 0x03,
    MF = 0x04,
    CNAME = 0x05,
    SOA = 0x06,
    MB = 0x07,
    MG = 0x08,
    MR = 0x09,
    NULL = 0x0a,
    WKS = 0x0b,
    PTR = 0x0c,
    HINFO = 0x0d,
    MINFO = 0x0e,
    MX = 0x0f,
    TXT = 0x10,
    AAAA = 0x1c
)

DNS_CLASSES = dict(
    IN = 0x01
)

DNS_TYPES_R = dict(zip(DNS_TYPES.values(), DNS_TYPES.keys()))
DNS_CLASSES_R = dict(zip(DNS_CLASSES.values(), DNS_CLASSES.keys()))

class DNSRequest(netius.Request):

    def __init__(self, name, type = "a", cls = "in", callback = None):
        netius.Request.__init__(self, callback = callback)
        self.name = name
        self.type = type
        self.cls = cls

    def request(self):

        format = "!HBBHHHH"

        result = []
        buffer = []

        first_flags = 0x0

        first_flags += DNS_QUERY << 7
        first_flags += DNS_SQUERY << 3
        first_flags |= DNS_RD

        second_flags = 0x0

        result.append(self.id)
        result.append(first_flags)
        result.append(second_flags)
        result.append(0x1)
        result.append(0x0)
        result.append(0x0)
        result.append(0x0)

        data = struct.pack(format, *result)
        buffer.append(data)

        query = self._query(
            self.name,
            type = self.type,
            cls = self.cls
        )
        buffer.append(query)

        data = b"".join(buffer)

        return data

    def _query(self, name, type = "a", cls = "in"):
        type_i = DNS_TYPES.get(type.upper(), 0x00)
        clsi = DNS_CLASSES.get(cls.upper(), 0x00)

        format = "!HH"

        data = self._label(name)
        data += struct.pack(format, type_i, clsi)

        return data

    def _label(self, value):
        buffer = []
        format = "!B"

        parts = value.split(".")
        for part in parts:
            part = netius.legacy.bytes(part)
            part_l = len(part)
            prefix = struct.pack("!B", part_l)
            part_s = prefix + part
            buffer.append(part_s)

        buffer.append(b"\0")
        data = b"".join(buffer)
        return data

class DNSResponse(netius.Response):

    def parse(self):

        format = "!HBBHHHH"

        self.header = self.data[:12]
        result = struct.unpack(format, self.header)

        self.id = result[0]
        self.first_flags = result[1]
        self.second_flags = result[2]
        self.qdcount = result[3]
        self.ancount = result[4]
        self.nscount = result[5]
        self.arcount = result[6]

        self.queries = []
        self.answers = []
        self.name_servers = []
        self.authorities = []

        index = 12

        for _index in range(self.qdcount):
            index, query = self.parse_qd(self.data, index)
            self.queries.append(query)

        for _index in range(self.ancount):
            index, answer = self.parse_an(self.data, index)
            self.answers.append(answer)

    def get_id(self):
        return self.id

    def parse_qd(self, data, index):
        index, name = self.parse_label(data, index)
        index, type = self.parse_short(data, index)
        index, cls = self.parse_short(data, index)
        type_s = DNS_TYPES_R.get(type, "undefined")
        cls_s = DNS_CLASSES_R.get(cls, "undefined")
        return (index, (name, type_s, cls_s))

    def parse_an(self, data, index):
        index, name = self.parse_label(data, index)
        index, type = self.parse_short(data, index)
        index, cls = self.parse_short(data, index)
        index, ttl = self.parse_long(data, index)
        index, _size = self.parse_short(data, index)
        type_s = DNS_TYPES_R.get(type, "undefined")
        cls_s = DNS_CLASSES_R.get(cls, "undefined")
        index, payload = self.parse_payload(data, index, type_s)
        return (index, (name, type_s, cls_s, ttl, payload))

    def parse_payload(self, data, index, type_s):
        type_s = type_s.lower()
        method_name = "parse_" + type_s
        method = getattr(self, method_name)
        return method(data, index)

    def parse_a(self, data, index):
        index, address = self.parse_ip4(data, index)
        return (index, address)

    def parse_aaaa(self, data, index):
        index, address = self.parse_ip6(data, index)
        return (index, address)

    def parse_mx(self, data, index):
        index, preference = self.parse_short(data, index)
        index, address = self.parse_label(data, index)
        return (index, (preference, address))

    def parse_cname(self, data, index):
        index, address = self.parse_label(data, index)
        return (index, address)

    def parse_ns(self, data, index):
        pass

    def parse_ar(self, data, index):
        pass

    def parse_label(self, data, index):
        buffer = []

        while True:
            initial = data[index]
            initial_i = netius.legacy.ord(initial)

            if initial_i == 0: index += 1; break
            is_pointer = initial_i & 0xc0

            if is_pointer:
                index, _data = self.parse_pointer(data, index)
                buffer.append(_data)
                data = b".".join(buffer)
                return (index, data)

            _data = data[index + 1:index + initial_i + 1]

            buffer.append(_data)
            index += initial_i + 1

        data = b".".join(buffer)
        return (index, data)

    def parse_pointer(self, data, index):
        slice = data[index:index + 2]

        offset, = struct.unpack("!H", slice)
        offset &= 0x3fff

        _index, label = self.parse_label(data, offset)

        return (index + 2, label)

    def parse_ip4(self, data, index):
        index, long = self.parse_long(data, index)
        address = netius.common.addr_to_ip4(long)
        return (index, address)

    def parse_ip6(self, data, index):
        index, long_long_first = self.parse_long_long(data, index)
        index, long_long_second = self.parse_long_long(data, index)
        address_i = (long_long_first << 64) + long_long_second
        address = netius.common.addr_to_ip6(address_i)
        return (index, address)

    def parse_byte(self, data, index):
        _data = data[index:index + 1]
        short, = struct.unpack("!B", _data)
        return (index + 1, short)

    def parse_short(self, data, index):
        _data = data[index:index + 2]
        short, = struct.unpack("!H", _data)
        return (index + 2, short)

    def parse_long(self, data, index):
        _data = data[index:index + 4]
        long, = struct.unpack("!L", _data)
        return (index + 4, long)

    def parse_long_long(self, data, index):
        _data = data[index:index + 8]
        long_long, = struct.unpack("!Q", _data)
        return (index + 8, long_long)

class DNSProtocol(netius.DatagramProtocol):

    ns_file_l = None

    @classmethod
    def ns_system(cls, type = "ip4"):
        ns = cls.ns_file(type = type)
        if ns: return ns[0]
        ns = cls.ns_google(type = type)
        if ns: return ns[0]
        ns = cls.ns_cloudfare(type = type)
        if ns: return ns[0]
        return None

    @classmethod
    def ns_file(cls, type = "ip4", force = False):
        # verifies if the list value for the file based name server
        # retrieval value is defined and if that's the case and the
        # force flag is not set returns it immediately
        if not cls.ns_file_l == None and not force: return cls.ns_file_l

        # verifies if the resolve file exists and if it does not returns
        # immediately indicating the impossibility to resolve the value
        if not os.path.exists("/etc/resolv.conf"): return None

        # retrieves the reference to the function that is going to validate
        # if the provided name server complies with the proper (address) type
        validator = getattr(netius.common, "is_" + type)

        # opens the resolve file and reads the complete set of contents
        # from it, closing the file afterwards
        file = open("/etc/resolv.conf", "rb")
        try: data = file.read()
        finally: file.close()

        # starts the list that is going to store the various name server
        # values, this is going to be populated with the file contents
        cls.ns_file_l = []

        # splits the contents of the file around their lines and tries
        # to find the name servers defined in it to be added to the list
        for line in data.split(b"\n"):
            line = line.strip()
            if not line.startswith(b"nameserver"): continue
            _header, ns = line.split(b" ", 1)
            ns = ns.strip()
            ns = netius.legacy.str(ns)
            is_valid = validator(ns)
            if not is_valid: continue
            cls.ns_file_l.append(ns)

        # returns the final value of the list of name servers loaded from
        # the file (as expected by the call)
        return cls.ns_file_l

    @classmethod
    def ns_google(cls, type = "ip4"):
        if type == "ip4": return ["8.8.8.8", "8.8.4.4"]
        if type == "ip6": return [
            "2001:4860:4860::8888",
            "2001:4860:4860::8844"
        ]
        return []

    @classmethod
    def ns_cloudfare(cls, type = "ip4"):
        if type == "ip4": return ["1.1.1.1", "1.0.0.1"]
        if type == "ip6": return [
            "2606:4700:4700::1111",
            "2606:4700:4700::1001"
        ]
        return []

    def query(self, name, type = "a", cls = "in", ns = None, callback = None):
        # retrieves the reference to the class associated with the
        # current instance to be used to access class operations
        _cls = self.__class__

        # verifies if a target name server was specified for the query
        # in case it was not uses the default (globally provided) value
        # that may be used for generic queries assuming internet access
        ns = ns or _cls.ns_system()

        # creates a new DNS request object describing the query that was
        # just sent and then generates the request stream code that is
        # going to be used for sending the request through network
        request = DNSRequest(
            name,
            type = type,
            cls = cls,
            callback = callback
        )
        data = request.request()

        # prints some debug information about the DNS query that is going
        # to be performed (provides some development capabilities)
        self.debug("Running DNS query %s '%s in '%s'" % (type, name, ns))

        # adds the current request pending callback handing to the internal
        # management structures so that it becomes callable latter
        self.add_request(request)

        # creates the final address assuming default port in the
        # name server and then send the contents of the DNS request
        address = (ns, 53)
        self.send(data, address)

    def on_data(self, address, data):
        netius.DatagramProtocol.on_data(self, address, data)

        # create the DNS response with the provided data stream and
        # runs the parse operation in it so that the response becomes
        # populated with the proper contents, this operation is risky as
        # it may fail in case the message is malformed
        response = DNSResponse(data)
        response.parse()

        # calls the DNS specific data handler with the proper response
        # object populated with the response from the DNS server
        self.on_data_dns(address, response)

    def on_data_dns(self, address, response):
        # tries to retrieve the request associated with the current
        # response and in case none is found returns immediately as
        # there's nothing remaining to be done
        request = self.get_request(response)
        if not request: return

        # removes the request being handled from the current request
        # structures so that a callback is no longer answered
        self.remove_request(request)

        # in case no callback is not defined for the request returns
        # immediately as there's nothing else remaining to be done,
        # otherwise calls the proper callback with the response
        if not request.callback: return
        request.callback(response)

class DNSClient(netius.ClientAgent):

    protocol = DNSProtocol

    @classmethod
    def query_s(
        cls,
        name,
        type = "a",
        cls_ = "in",
        ns = None,
        callback = None,
        loop = None
    ):
        ns = ns or cls.protocol.ns_system()
        address = (ns, 53)
        protocol = cls.protocol()

        def on_connect(result):
            _transport, protocol = result
            protocol.query(
                name,
                type = type,
                cls = cls_,
                ns = ns,
                callback = callback
            )

        loop = netius.build_datagram(
            lambda: protocol,
            callback = on_connect,
            loop = loop,
            remote_addr = address
        )

        return loop, protocol

if __name__ == "__main__":
    def handler(response):
        # closes the current protocol to correctly close
        # all of the underlying structures
        protocol.close()

        # retrieves the currently associated loop using
        # netius base infra-structure and then runs the
        # stop operation on the next tick end
        netius.compat_loop(loop).stop()

        # in case the provided response is not valid
        # a timeout message is printed to indicate the
        # problem with the resolution
        if not response: print("Timeout in resolution"); return

        # unpacks the complete set of contents from
        # the various answers so that only the address
        # of the answer is available, then prints them
        for answer in response.answers:
            if type in ("a", "aaaa"):
                address = answer[4]
                print("%s" % address)
            if type in ("mx",):
                extra = answer[4]
                priority = extra[0]
                address = extra[1]
                print("%s => %d" % (address, priority))

    # retrieves the values of the configuration variables
    # that are going to be used to perform the DNS query
    name = netius.conf("DNS_NAME", "gmail.com")
    type = netius.conf("DNS_TYPE", "mx")
    ns = netius.conf("DNS_NS", None)

    # runs the static version of a DNS query, note that
    # the daemon flag is unset so that the global client
    # runs in foreground avoiding the exit of the process
    loop, protocol = DNSClient.query_s(
        name,
        type = type,
        ns = ns,
        callback = handler
    )
    loop.run_forever()
    loop.close()
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os

CA_URL = "https://curl.haxx.se/ca/cacert.pem"

COMMON_PATH = os.path.dirname(__file__)
BASE_PATH = os.path.join(COMMON_PATH, "..", "base")
EXTRAS_PATH = os.path.join(BASE_PATH, "extras")
SSL_CA_PATH = os.path.join(EXTRAS_PATH, "net.ca")

def ensure_setup():
    ensure_ca()

def ensure_ca(path = SSL_CA_PATH):
    if os.path.exists(path): return
    _download_ca(path = path)

def _download_ca(path = SSL_CA_PATH):
    import netius.clients
    result = netius.clients.HTTPClient.method_s(
        "GET",
        CA_URL,
        asynchronous = False
    )
    response = netius.clients.HTTPClient.to_response(result)
    contents = response.read()
    _store_contents(contents, path)

def _store_contents(contents, path):
    file = open(path, "wb")
    try: file.write(contents)
    finally: file.close()
    return path



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from . import observer

OPEN = 1
""" The open status value, meant to be used in situations
where the status of the entity is open (opposite of d) """

CLOSED = 2
""" Closed status value to be used in entities which have
no pending structured opened and operations are limited """

PENDING = 3
""" The pending status used for transient states (eg created)
connections under this state must be used carefully """

class Stream(observer.Observable):
    """
    Abstract stream class responsible for the representation of
    a "virtual" connection state for situation where multiplexing
    of single connection exists.

    Most of the interface for a stream should be "logically" similar
    to the one defined by a connection.
    """

    def __init__(self, owner = None):
        observer.Observable.__init__(self)
        self.status = PENDING
        self.owner = owner
        self.connection = owner.owner

    def reset(self):
        pass

    def open(self):
        if self.status == OPEN: return
        self.status = OPEN
        self.connection.owner.on_stream_c(self)

    def close(self):
        if self.status == CLOSED: return
        self.status = CLOSED
        self.connection.owner.on_stream_d(self)

    def info_dict(self, full = False):
        info = dict(
            status = self.status
        )
        return info

    def is_open(self):
        return self.status == OPEN

    def is_closed(self):
        return self.status == CLOSED

    def is_pending(self):
        return self.status == PENDING



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import time

from . import errors

REQUEST_TIMEOUT = 10.0
""" The timeout until a request is considered to be
expired and is discarded from the request related
structures, this is crucial to avoid memory leaks """

class Request(object):
    """
    Abstract request structure used to represent
    a request in a server/client model, this allows
    for easy identification and response (callback).
    """

    IDENTIFIER = 0x0000
    """ The global class identifier value that is going to
    be used when assigning new values to the request """

    def __init__(self, timeout = REQUEST_TIMEOUT, callback = None):
        self.id = self.__class__._generate_id()
        self.timeout = time.time() + timeout
        self.callback = callback

    @classmethod
    def _generate_id(cls):
        cls.IDENTIFIER = (cls.IDENTIFIER + 1) & 0xffff
        return cls.IDENTIFIER

class Response(object):
    """
    Top level abstract representation of a response to
    be sent based on a previously created request, the
    input of this object should be raw data and a relation
    between the request and the response is required.

    The association/relation between the response and the
    request should be done using the original request
    generated identifier.
    """

    def __init__(self, data, request = None):
        self.data = data
        self.request = request

    def parse(self):
        pass

    def get_request(self):
        return self.request

    def get_id(self):
        raise errors.NetiusError("Not implemented")



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import ssl
import struct

import netius.common

from . import http

class HTTP2Connection(http.HTTPConnection):

    def __init__(
        self,
        legacy = True,
        window = netius.common.HTTP2_WINDOW,
        settings = netius.common.HTTP2_SETTINGS_OPTIMAL,
        settings_r = netius.common.HTTP2_SETTINGS,
        *args,
        **kwargs
    ):
        http.HTTPConnection.__init__(self, *args, **kwargs)
        self.legacy = legacy
        self.settings = dict(settings)
        self.settings_r = dict(settings_r)
        self.window = window
        self.window_o = self.settings[netius.common.http2.SETTINGS_INITIAL_WINDOW_SIZE]
        self.window_l = self.window_o
        self.window_t = self.window_o // 2
        self.preface = False
        self.preface_b = b""
        self.frames = []
        self.unavailable = {}

    def open(self, *args, **kwargs):
        http.HTTPConnection.open(self, *args, **kwargs)
        if not self.is_open(): return
        if not self.legacy: self.set_h2()

    def info_dict(self, full = False):
        info = http.HTTPConnection.info_dict(self, full = full)
        info.update(
            legacy = self.legacy,
            window = self.window,
            window_o = self.window_o,
            window_l = self.window_l,
            window_t = self.window_t,
            frames = len(self.frames)
        )
        return info

    def flush_s(self, stream = None, callback = None):
        return self.send_part(
            b"",
            stream = stream,
            final = True,
            flush = True,
            callback = callback
        )

    def set_h2(self):
        self.legacy = False
        if self.parser: self.parser.destroy()
        self.parser = netius.common.HTTP2Parser(self, store = True)
        self.parser.bind("on_data", self.on_data)
        self.parser.bind("on_header", self.on_header)
        self.parser.bind("on_payload", self.on_payload)
        self.parser.bind("on_frame", self.on_frame)
        self.parser.bind("on_data_h2", self.on_data_h2)
        self.parser.bind("on_headers_h2", self.on_headers_h2)
        self.parser.bind("on_rst_stream", self.on_rst_stream)
        self.parser.bind("on_settings", self.on_settings)
        self.parser.bind("on_ping", self.on_ping)
        self.parser.bind("on_goaway", self.on_goaway)
        self.parser.bind("on_window_update", self.on_window_update)
        self.parser.bind("on_continuation", self.on_continuation)

    def parse(self, data):
        if not self.legacy and not self.preface:
            data = self.parse_preface(data)
            if not data: return
        return self.parser.parse(data)

    def parse_preface(self, data):
        """
        Tries to run the parsing on the preface part of the
        connection establishment using the provided data
        note that the data is buffered in case the proper size
        has not been reached for proper validation.

        This should be the first step when trying to establish
        a proper HTTP 2 connection.

        :type data: String
        :param data: The data buffer that is going to be used to
        try to parse the connection preface.
        :rtype: String
        :return: The resulting data after the preface has been
        parsed, this should be empty or invalid in case no data
        is pending to be parsed.
        """

        # adds the current data to the buffer of bytes pending
        # in the preface parsing and then verified that the proper
        # preface size has been reached, in case it has not returned
        # an invalid value immediately (no further parsing)
        self.preface_b += data
        preface_l = len(netius.common.HTTP2_PREFACE)
        is_size = len(self.preface_b) >= preface_l
        if not is_size: return None

        # retrieves the preface string from the buffer (according to size)
        # and runs the string based verification, raising an exception in
        # case there's a mismatch in the string validation
        preface = self.preface_b[:preface_l]
        if not preface == netius.common.HTTP2_PREFACE:
            raise netius.ParserError("Invalid preface")

        # sets the preface (parsed) flag indicating that the preface has
        # been parsed for the current connection
        self.preface = True

        # retrieves the extra data added to the preface buffer and then
        # unsets the same buffer (no more preface parsing)
        data = self.preface_b[preface_l:]
        self.preface_b = b""

        # calls the proper callback for the preface sending both the current
        # instance and the associated parser for handling
        self.owner.on_preface_http2(self, self.parser)

        # returns the remaining data pending to be parsed so that it may be
        # parsed by any extra operation
        return data

    def send_plain(
        self,
        data,
        stream = None,
        final = True,
        delay = True,
        callback = None
    ):
        if self.legacy: return http.HTTPConnection.send_plain(
            self,
            data,
            stream = stream,
            final = final,
            delay = delay,
            callback = callback
        )

        # verifies if the data should be fragmented for the provided
        # stream and if that's not required send the required data
        # straight away with any required splitting/fragmentation of it
        if not self.fragmentable_stream(stream, data):
            return self.send_data(
                data,
                stream = stream,
                end_stream = final,
                delay = delay,
                callback = callback
            )

        # sends the same data but using a fragmented approach where the
        # data is going to be splitted according to the maximum determined
        # frame size, this is required to overcome limitations in the connection
        # that has been established with the other peer
        return self.send_fragmented(
            data,
            stream = stream,
            final = final,
            delay = delay,
            callback = callback
        )

    def send_chunked(
        self,
        data,
        stream = None,
        final = True,
        delay = True,
        callback = None
    ):
        if self.legacy: return http.HTTPConnection.send_chunked(
            self,
            data,
            stream = stream,
            final = final,
            delay = delay,
            callback = callback
        )
        return self.send_plain(
            data,
            stream = stream,
            final = final,
            delay = delay,
            callback = callback
        )

    def send_fragmented(
        self,
        data,
        stream = None,
        final = True,
        delay = True,
        callback = None
    ):
        count = 0
        fragments = self.fragment_stream(stream, data)
        fragments = list(fragments)
        fragments_l = len(fragments)

        for index in netius.legacy.xrange(fragments_l):
            is_last = index == fragments_l - 1
            fragment = fragments[index]
            if is_last:
                count += self.send_data(
                    fragment,
                    stream = stream,
                    end_stream = final,
                    delay = delay,
                    callback = callback
                )
            else:
                count += self.send_data(
                    fragment,
                    stream = stream,
                    end_stream = False,
                    delay = delay
                )

        return count

    def send_response(
        self,
        data = None,
        headers = None,
        version = None,
        code = 200,
        code_s = None,
        apply = False,
        stream = None,
        final = True,
        flush = True,
        delay = True,
        callback = None
    ):
        # in case the legacy mode is enabled the send response call is
        # forwarded to the upper layers so that it's handled properly
        if self.legacy: return http.HTTPConnection.send_response(
            self,
            data = data,
            headers = headers,
            version = version,
            code = code,
            code_s = code_s,
            apply = apply,
            stream = stream,
            final = final,
            flush = flush,
            delay = delay,
            callback = callback
        )

        # retrieves the various parts that define the response
        # and runs a series of normalization processes to retrieve
        # the relevant information of the data to be sent to client
        data = data or b""
        data = netius.legacy.bytes(data)
        headers = headers or dict()
        data_l = len(data) if data else 0
        is_empty = code in (204, 304) and data_l == 0

        # runs a series of verifications taking into account the type
        # of the method defined in the current request, for instance if
        # the current request is a HEAD one then no data is sent (as expected)
        if self.parser_ctx.method and self.parser_ctx.method.upper() == "HEAD":
            data = b""

        # verifies if the content length header is currently present
        # in the provided headers and in case it's not inserts it
        if not "content-length" in headers and not is_empty:
            headers["content-length"] = str(data_l)

        # in case the apply flag is set the apply all operation is performed
        # so that a series of headers are applied to the current context
        # (things like the name of the server connection, etc)
        if apply: self.owner._apply_all(self.parser, self, headers)

        # sends the initial headers data (including status line), this should
        # trigger the initial data sent to the peer/client
        count = self.send_header(
            headers = headers,
            version = version,
            code = code,
            code_s = code_s,
            stream = stream
        )

        # sends the part/payload information (data) to the client and optionally
        # flushes the current internal buffers to enforce sending of the value
        count += self.send_part(
            data,
            stream = stream,
            final = final,
            flush = flush,
            delay = delay,
            callback = callback
        )
        return count

    def send_header(
        self,
        headers = None,
        version = None,
        code = 200,
        code_s = None,
        stream = None,
        final = False,
        delay = True,
        callback = None
    ):
        # in case the legacy mode is enabled the send header call is
        # forwarded to the upper layers so that it's handled properly
        if self.legacy: return http.HTTPConnection.send_header(
            self,
            headers = headers,
            version = version,
            code = code,
            code_s = code_s,
            stream = stream,
            delay = delay,
            callback = callback
        )

        # verifies if the headers value has been provided and in case it
        # has not creates a new empty dictionary (runtime compatibility)
        headers = headers or dict()

        # defines the proper default base http version in case it has not
        # been provided as part the default values
        version = version or "HTTP/2.0"

        # creates the headers base list that is going to store the various
        # header tuples representing the headers in canonical http2 form
        headers_b = []
        headers_b.append((":status", str(code)))

        # iterates over the complete set of raw header values to normalize
        # them and add them to the currently defined base list
        for key, value in netius.legacy.iteritems(headers):
            key = netius.common.header_down(key)
            if key in ("connection", "transfer-encoding"): continue
            if not isinstance(value, list): value = (value,)
            for _value in value: headers_b.append((key, _value))

        # verifies if this is considered to be the final operation in the stream
        # and if that's the case creates a new callback for the closing of the
        # stream at the end of the operation, this is required for proper collection
        if final:
            old_callback = callback

            def callback(connection):
                self.close_stream(stream, final = final)
                old_callback and old_callback(connection)

        # runs the send headers operations that should send the headers list
        # to the other peer and returns the number of bytes sent
        count = self.send_headers(
            headers_b,
            end_stream = final,
            stream = stream,
            delay = delay,
            callback = callback
        )

        # "notifies" the owner of the connection that the headers have been
        # sent all the http header information should be present
        self.owner.on_send_http(
            self.connection_ctx,
            self.parser_ctx,
            headers = headers,
            version = version,
            code = code,
            code_s = code_s
        )

        # returns the final number of bytes that have been sent during the current
        # operation of sending headers to the other peer
        return count

    def send_part(
        self,
        data,
        stream = None,
        final = True,
        flush = False,
        delay = True,
        callback = None
    ):
        if self.legacy: return http.HTTPConnection.send_part(
            self,
            data,
            stream = stream,
            final = final,
            flush = flush,
            delay = delay,
            callback = callback
        )

        # verifies if this is considered to be the final operation in the stream
        # and if that's the case creates a new callback for the closing of the
        # stream at the end of the operation, this is required for proper collection
        if final:
            old_callback = callback

            def callback(connection):
                self.close_stream(stream, final = final)
                old_callback and old_callback(connection)

        # verifies if the current connection/stream is flushed meaning that it requires
        # a final chunk of data to be sent to the peer, if that's not the case there's
        # no need to run the flushing as a possible empty data frame may be sent which
        # may cause errors to be raised from the server side
        flush = flush and self.is_flushed()

        if flush:
            count = self.send_base(
                data,
                stream = stream,
                final = False
            )
            self.flush(stream = stream, callback = callback)
        else:
            count = self.send_base(
                data,
                stream = stream,
                final = final,
                delay = delay,
                callback = callback
            )
        return count

    def send_frame(
        self,
        type = 0x01,
        flags = 0x00,
        payload = b"",
        stream = 0x00,
        delay = True,
        callback = None
    ):
        size = len(payload)
        size_h = size >> 16
        size_l = size & 0xffff
        header = struct.pack("!BHBBI", size_h, size_l, type, flags, stream)
        message = header + payload
        self.owner.on_send_http2(self, self.parser, type, flags, payload, stream)
        return self.send(message, delay = delay, callback = callback)

    def send_data(
        self,
        data = b"",
        end_stream = True,
        stream = None,
        delay = True,
        callback = None
    ):
        # builds the flags byte taking into account the various
        # options that have been passed to the sending of data
        flags = 0x00
        data_l = len(data)
        if end_stream: flags |= 0x01

        # builds the callback clojure so that the connection state
        # is properly updated upon the sending of data
        callback = self._build_c(callback, stream, data_l)

        # verifies if the stream is available for the amount of data
        # that is currently being sent and if that's not the case delays
        # the sending of the frame to when the stream becomes available
        if not self.available_stream(stream, data_l):
            count = self.delay_frame(
                type = netius.common.DATA,
                flags = flags,
                payload = data,
                stream = stream,
                delay = delay,
                callback = callback
            )
            self.try_unavailable(stream)
            return count

        # runs the increments remove window value, decrementing the window
        # by the size of the data being sent
        self.increment_remote(stream, data_l * -1, all = True)

        # runs the "proper" sending of the data frame, registering the callback
        # with the expected clojure
        count = self.send_frame(
            type = netius.common.DATA,
            flags = flags,
            payload = data,
            stream = stream,
            delay = delay,
            callback = callback
        )

        # runs the try unavailable method to verify if the stream did became
        # unavailable after the sending of the data
        self.try_unavailable(stream)

        # returns the final number of bytes sent to the called method, this should
        # match the value of the data length
        return count

    def send_headers(
        self,
        headers = [],
        end_stream = False,
        end_headers = True,
        stream = None,
        delay = True,
        callback = None
    ):
        flags = 0x00
        if end_stream: flags |= 0x01
        if end_headers: flags |= 0x04
        payload = self.parser.encoder.encode(headers)
        return self.send_frame(
            type = netius.common.HEADERS,
            flags = flags,
            payload = payload,
            stream = stream,
            delay = delay,
            callback = callback
        )

    def send_rst_stream(
        self,
        error_code = 0x00,
        stream = None,
        delay = True,
        callback = None
    ):
        payload = struct.pack("!I", error_code)
        return self.send_frame(
            type = netius.common.RST_STREAM,
            payload = payload,
            stream = stream,
            delay = delay,
            callback = callback
        )

    def send_settings(
        self,
        settings = (),
        ack = False,
        delay = True,
        callback = None
    ):
        flags = 0x00
        if ack: flags |= 0x01
        buffer = []
        for ident, value in settings:
            setting_s = struct.pack("!HI", ident, value)
            buffer.append(setting_s)
        payload = b"".join(buffer)
        return self.send_frame(
            type = netius.common.SETTINGS,
            flags = flags,
            payload = payload,
            delay = delay,
            callback = callback
        )

    def send_ping(
        self,
        opaque = b"\0\0\0\0\0\0\0\0",
        ack = False,
        delay = True,
        callback = None
    ):
        flags = 0x00
        if ack: flags |= 0x01
        return self.send_frame(
            type = netius.common.PING,
            flags = flags,
            payload = opaque,
            delay = delay,
            callback = callback
        )

    def send_goaway(
        self,
        last_stream = 0x00,
        error_code = 0x00,
        message = "",
        close = True,
        delay = True,
        callback = None
    ):
        if close:
            old_callback = callback

            def callback(connection):
                self.close()
                old_callback and old_callback(connection)

        message = netius.legacy.bytes(message)
        payload = struct.pack("!II", last_stream, error_code)
        payload += message
        return self.send_frame(
            type = netius.common.GOAWAY,
            payload = payload,
            delay = delay,
            callback = callback
        )

    def send_window_update(
        self,
        increment = 0,
        stream = None,
        delay = True,
        callback = None
    ):
        payload = struct.pack("!I", increment)
        return self.send_frame(
            type = netius.common.WINDOW_UPDATE,
            payload = payload,
            stream = stream,
            delay = delay,
            callback = callback
        )

    def send_delta(self):
        delta = self.window_l -\
            netius.common.HTTP2_SETTINGS[netius.common.http2.SETTINGS_INITIAL_WINDOW_SIZE]
        if delta == 0: return
        self.send_window_update(increment = delta, stream = 0x00)

    def delay_frame(self, *args, **kwargs):
        # retrieves the reference to the stream identifier for which
        # the frame is meant to be sent, and then uses this same value
        # to try to retrieve the target stream of the frame
        stream = kwargs["stream"]
        stream = self.parser._get_stream(stream)

        # adds the frame structure (tuple) as the structure describing
        # the frame to be delayed, then increments the frame counter in
        # the stream so that it represent a proper value
        self.frames.append((args, kwargs))
        stream.frames += 1

        # returns a zero value indicating that no bytes have been sent
        # "immediately" by this method
        return 0

    def flush_frames(self, all = True):
        """
        Runs the flush operation on the delayed/pending frames, meaning
        that the window/availability tests are going to be run, checking
        if the various streams and connection are ready for sending the
        frames.

        In case the all flag is active the complete set of frames are going
        to be tested for sending, this operation implies more resource usage.

        This method should be called after a window update frame is
        received so that the pending frames may be sent.

        :type all: bool
        :param all: If the complete set of frames should be tested, or
        if instead at the first testing fail the control flow should be
        returned immediately.
        :rtype: bool
        :return: If all the pending frames have been successfully flushed.
        """

        # starts the values for both the offset value to be used in the
        # pop operation and the dictionary to be used in the storage of
        # the bitset of streams marked as started in the iteration
        offset = 0
        starved = dict() if all else None

        # iterates over the complete set of frames pending to to be sent
        # (delayed) trying to send each of them until one fails and the
        # flushing operation is delayed until further requesting
        while offset < len(self.frames):
            # retrieves the reference to the current frame tuple to
            # be sent and retrieves the stream and payload from it
            frame = self.frames[offset]
            args, kwargs = frame
            stream = kwargs["stream"]
            payload = kwargs["payload"]
            payload_l = len(payload)

            # verifies that the stream is currently place in the list of
            # stream that are considered unavailable as this is a state
            # required for proper execution
            netius.verify(stream in self.unavailable)

            # verifies if the stream associated with the frame to be
            # sent is in the started map and if that's the case continue
            # the current loop immediately (cannot flush frame)
            if starved and stream in starved:
                offset += 1
                continue

            # retrieves the reference to the stream object from the
            # identifier of the stream, this may an invalid/unset value
            _stream = self.parser._get_stream(stream, strict = False)

            # verifies if the current stream to be flushed is still
            # open and if that's not the case removes the frame from
            # the frames queue and skips the current iteration
            if not _stream or not _stream.is_open():
                self.frames.pop(offset)
                if _stream: _stream.frames -= 1
                continue

            # makes sure that the stream is currently marked as not available
            # this should be the state for every stream that has pending frames
            netius.verify(not _stream._available)

            # verifies if there's available "space" in the stream flow
            # to send the current payload and in case there's not breaks
            # the current loop as there's nothing else to be done, delays
            # pending frames for a new flush operation, note that the
            # return value is invalid (meaning that the stream may not
            # be available), a special failover operation exists if the
            # all flush operation is enabled in which the stream is marked
            # as starved and the current iteration is skipped trying to
            # flush frames from different streams
            available = self.available_stream(stream, payload_l, strict = False)
            if not available and not all: return False
            if not available and all:
                starved[stream] = True
                offset += 1
                continue

            # removes the frame from both of the frame queues (both global
            # and stream) so that it is no longer going to be used for flush
            self.frames.pop(offset)
            _stream.frames -= 1

            # decrements the current stream window by the size of the payload
            # and then runs the send frame operation for the pending frame
            self.increment_remote(stream, payload_l * -1, all = True)
            self.send_frame(*args, **kwargs)

        # returns the final result with a valid value meaning that all of the
        # flush operations have been successful (no frames pending in connection)
        return True if offset == 0 else False

    def flush_available(self):
        """
        Runs the (became) available flush operation that tries to determine
        all the streams that were under the "blocked" state and became
        "unblocked", notifying them about that "edge" operation.

        This operation must be performed after any of the blocking constraints
        is changed (eg: connection window, stream window, etc.).
        """

        # iterates over the complete set of streams (identifiers) that are
        # currently under the unavailable/blocked state, to try to determine
        # if they became unblocked by the "current operation"
        for stream in netius.legacy.keys(self.unavailable):
            self.try_available(stream)

    def set_settings(self, settings):
        self.settings_r.update(settings)

    def close_stream(self, stream, final = False, flush = False, reset = False):
        if not self.parser._has_stream(stream): return
        stream = self.parser._get_stream(stream)
        if not stream: return
        stream.end_stream_l = final
        stream.close(flush = flush, reset = reset)

    def available_stream(self, stream, length, strict = True):
        if self.window == 0: return False
        if self.window < length: return False
        stream = self.parser._get_stream(stream)
        if not stream: return True
        if stream.window == 0: return False
        if stream.window < length: return False
        if strict and stream.frames: return False
        return True

    def fragment_stream(self, stream, data):
        stream = self.parser._get_stream(stream)
        return stream.fragment(data)

    def fragmentable_stream(self, stream, data):
        stream = self.parser._get_stream(stream)
        return stream.fragmentable(data)

    def open_stream(self, stream):
        stream = self.parser._get_stream(stream, strict = False)
        if not stream : return False
        return True if stream and stream.is_open() else False

    def try_available(self, stream, strict = True):
        """
        Tries to determine if the stream with the provided identifier
        has just became available (unblocked from blocked state), this
        happens when the required window value (either connection or
        stream is increased properly).

        :type stream: int
        :param stream: The identifier of the stream that is going to
        be tested from proper connection availability.
        :type strict: bool
        :param strict: If the strict mode should be used in the availability
        testing, this implies extra verifications.
        """

        # verifies if the stream is currently present in the map of unavailable
        # or blocked streams and if that's the case returns immediately as
        # the connection is not blocked
        if not stream in self.unavailable: return

        # tries to retrieve the stream object reference from the identifier and
        # in case none is retrieved (probably stream closed) returns immediately
        # and removes the stream from the map of unavailability
        _stream = self.parser._get_stream(stream, strict = False)
        if not _stream:
            del self.unavailable[stream]
            return

        # tries to determine if the stream is available for the sending of at
        # least one byte and if that's not the case returns immediately, not
        # setting the stream as available
        if not self.available_stream(stream, 1, strict = strict): return

        # removes the stream from the map of unavailable stream and "notifies"
        # the stream about the state changing operation to available/unblocked
        del self.unavailable[stream]
        _stream.available()

    def try_unavailable(self, stream, strict = True):
        """
        Runs the unavailability test on the stream with the provided identifier
        meaning that a series of validation will be performed to try to determine
        if for some reason is not possible to send any more data frames to the
        stream until some window changes. A stream that is under the unavailable
        state is considered "blocked".

        :type stream: int
        :param stream: The identifier of the stream that is going to
        be tested from proper connection unavailability.
        :type strict: bool
        :param strict: If the strict mode should be used in the availability
        testing, this implies extra verifications.
        """

        # in case the stream identifier is already present in the unavailable
        # map it cannot be marked as unavailable again
        if stream in self.unavailable: return

        # tries to retrieve the reference to the stream object to be tested
        # an in case none is found (connection closed) returns immediately
        _stream = self.parser._get_stream(stream, strict = False)
        if not _stream: return

        # runs the proper availability verification by testing the capacity
        # of the stream to send one byte and in case there's capacity to send
        # that byte the stream is considered available or unblocked, so the
        # control flow must be returned (stream not marked)
        if self.available_stream(stream, 1, strict = strict): return

        # marks the stream as unavailable and "notifies" the stream object
        # about the changing to the unavailable/blocked state
        self.unavailable[stream] = True
        _stream.unavailable()

    def increment_remote(self, stream, increment, all = False):
        """
        Increments the size of the remove window associated with
        the stream passed by argument by the size defined in the
        increment field (in bytes).

        If the stream is not provided or invalid the global window
        is updated instead of the stream one.

        :type stream: int
        :param stream: The identifier of the stream that is going
        to have its window incremented, or invalid if the global
        connection window is meant to be updated.
        :type increment: int
        :param increment: The increment in bytes for the window,
        this value may be negative for decrement operations.
        :type all: bool
        :param all: If all the resources (connection and stream)
        should be updated by the increment operation.
        """

        if not stream or all: self.window += increment
        if not stream: return
        stream = self.parser._get_stream(stream)
        if not stream: return
        stream.remote_update(increment)

    def increment_local(self, stream, increment):
        # increments the global connection local window
        # by the provided value, and then verifies if the
        # threshold has been passed, if that's the case
        # the window updated frame must be sent so that
        # the remove end point is properly notified
        self.window_l += increment
        if self.window_l < self.window_t:
            self.send_window_update(
                increment = self.window_o - self.window_l,
                stream = 0x00
            )
            self.window_l = self.window_o

        # tries to retrieve the stream associates with the
        # provided identifier and then runs the local update
        # operation in it (may trigger window update flushing)
        stream = self.parser._get_stream(stream)
        if not stream: return
        stream.local_update(increment)

    def error_connection(
        self,
        last_stream = 0x00,
        error_code = 0x00,
        message = "",
        close = True,
        callback = None
    ):
        self.send_goaway(
            last_stream = last_stream,
            error_code = error_code,
            message = message,
            close = close,
            callback = callback
        )

    def error_stream(
        self,
        stream,
        last_stream = 0x00,
        error_code = 0x00,
        message = "",
        close = True,
        callback = None
    ):
        self.send_rst_stream(
            error_code = error_code,
            stream = stream,
            callback = lambda c: self.error_connection(
                last_stream = last_stream,
                error_code = error_code,
                message = message,
                close = close,
                callback = callback
            )
        )

    def on_header(self, header):
        self.owner.on_header_http2(self, self.parser, header)

    def on_payload(self):
        self.owner.on_payload_http2(self, self.parser)

    def on_frame(self):
        self.owner.on_frame_http2(self, self.parser)

    def on_data_h2(self, stream, contents):
        self.increment_local(
            stream and stream.identifier,
            increment = len(contents) * -1
        )
        self.owner.on_data_http2(self, self.parser, stream, contents)

    def on_headers_h2(self, stream):
        self.owner.on_headers_http2(self, self.parser, stream)

    def on_rst_stream(self, stream, error_code):
        self.owner.on_rst_stream_http2(self, self.parser, stream, error_code)

    def on_settings(self, settings, ack):
        self.owner.on_settings_http2(self, self.parser, settings, ack)

    def on_ping(self, opaque, ack):
        self.owner.on_ping_http2(self, self.parser, opaque, ack)

    def on_goaway(self, last_stream, error_code, extra):
        self.owner.on_goaway_http2(self, self.parser, last_stream, error_code, extra)

    def on_window_update(self, stream, increment):
        self.increment_remote(stream and stream.identifier, increment)
        self.flush_frames()
        self.flush_available()
        self.owner.on_window_update_http2(self, self.parser, stream, increment)

    def on_continuation(self, stream):
        self.owner.on_continuation_http2(self, self.parser, stream)

    def is_throttleable(self):
        if self.legacy: return http.HTTPConnection.is_throttleable(self)
        return False

    @property
    def connection_ctx(self):
        if self.legacy: return super(HTTP2Connection, self).connection_ctx
        if not self.parser: return self
        if not self.parser.stream_o: return self
        return self.parser.stream_o

    @property
    def parser_ctx(self):
        if self.legacy: return super(HTTP2Connection, self).parser_ctx
        if not self.parser: return None
        if not self.parser.stream_o: return self.parser
        return self.parser.stream_o

    def _build_c(self, callback, stream, data_l):
        stream = self.parser._get_stream(stream, strict = False)
        if not stream: return callback

        stream.pending_s += data_l
        old_callback = callback

        def callback(connection):
            stream.pending_s -= data_l
            if not old_callback: return
            return old_callback(connection)

        return callback

    def _flush_plain(self, stream = None, callback = None):
        self.send_part(b"", stream = stream, callback = callback)

    def _flush_chunked(self, stream = None, callback = None):
        if self.legacy: return http.HTTPConnection._flush_chunked(
            self,
            stream = stream,
            callback = callback
        )
        self._flush_plain(stream = stream, callback = callback)

class HTTP2Server(http.HTTPServer):

    def __init__(
        self,
        legacy = True,
        safe = False,
        settings = netius.common.HTTP2_SETTINGS_OPTIMAL,
        *args,
        **kwargs
    ):
        self.legacy = legacy
        self.safe = safe
        self.settings = settings
        self.settings_t = netius.legacy.items(self.settings)
        self.has_h2 = self._has_h2()
        self.has_all_h2 = self._has_all_h2()
        self._protocols = []
        self.safe = self.get_env("SAFE", self.safe, cast = bool)
        http.HTTPServer.__init__(self, *args, **kwargs)

    @classmethod
    def _has_hpack(cls):
        try: import hpack #@UnusedImport
        except ImportError: return False
        return True

    @classmethod
    def _has_alpn(cls):
        return ssl.HAS_ALPN

    @classmethod
    def _has_npn(cls):
        return ssl.HAS_NPN

    def info_dict(self, full = False):
        info = http.HTTPServer.info_dict(self, full = full)
        info.update(
            legacy = self.legacy,
            safe = self.safe,
            has_h2 = self.has_h2,
            has_all_h2 = self.has_all_h2
        )
        return info

    def get_protocols(self):
        if self._protocols: return self._protocols
        if not self.safe and self.has_h2: self._protocols.extend(["h2"])
        if self.legacy: self._protocols.extend(["http/1.1", "http/1.0"])
        return self._protocols

    def build_connection(self, socket, address, ssl = False):
        return HTTP2Connection(
            owner = self,
            socket = socket,
            address = address,
            ssl = ssl,
            encoding = self.encoding,
            legacy = self.legacy,
            settings = self.settings
        )

    def on_exception(self, exception, connection):
        if hasattr(connection, "legacy") and connection.legacy:
            return http.HTTPServer.on_exception(self, exception, connection)
        if not isinstance(exception, netius.NetiusError):
            return http.HTTPServer.on_exception(self, exception, connection)
        try: self._handle_exception(exception, connection)
        except Exception: connection.close()

    def on_ssl(self, connection):
        http.HTTPServer.on_ssl(self, connection)
        if self.safe or not self.has_h2: return
        protocol = connection.ssl_protocol()
        if not protocol == "h2": return
        connection.set_h2()

    def on_serve(self):
        http.HTTPServer.on_serve(self)
        safe_s = "with" if self.safe else "without"
        self.info("Starting HTTP2 server %s safe mode ..." % safe_s)
        if not self.has_h2: self.info("No support for HTTP2 is available ...")
        elif not self.has_all_h2: self.info("Limited support for HTTP2 is available ...")
        for setting, name in netius.common.HTTP2_TUPLES:
            if not self.env: continue
            value = self.get_env(name, None, cast = int)
            if value == None: continue
            self.settings[setting] = value
            self.info("Setting HTTP2 setting %s with value '%d' ..." % (name, value))
        self.settings_t = netius.legacy.items(self.settings)

    def on_preface_http2(self, connection, parser):
        connection.send_settings(settings = self.settings_t)
        connection.send_delta()

    def on_header_http2(self, connection, parser, header):
        pass

    def on_payload_http2(self, connection, parser):
        is_debug = self.is_debug()
        is_debug and self._log_frame(connection, parser)

    def on_frame_http2(self, connection, parser):
        pass

    def on_data_http2(self, connection, parser, stream, contents):
        pass

    def on_headers_http2(self, connection, parser, stream):
        pass

    def on_rst_stream_http2(self, connection, parser, stream, error_code):
        if not stream: return
        stream.end_stream = True
        stream.end_stream_l = True
        stream.close(reset = False)

    def on_settings_http2(self, connection, parser, settings, ack):
        if ack: return
        self.debug("Received settings %s for connection" % str(settings))
        connection.set_settings(dict(settings))
        connection.send_settings(ack = True)

    def on_ping_http2(self, connection, parser, opaque, ack):
        if ack: return
        connection.send_ping(opaque = opaque, ack = True)

    def on_goaway_http2(self, connection, parser, last_stream, error_code, extra):
        if error_code == 0x00: return
        self._log_error(error_code, extra)

    def on_window_update_http2(self, connection, parser, stream, increment):
        self.debug("Window updated with increment %d bytes" % increment)

    def on_continuation_http2(self, connection, parser, stream):
        pass

    def on_send_http2(self, connection, parser, type, flags, payload, stream):
        is_debug = self.is_debug()
        is_debug and self._log_send(connection, parser, type, flags, payload, stream)

    def _has_h2(self):
        cls = self.__class__
        if not cls._has_hpack(): return False
        return True

    def _has_all_h2(self):
        cls = self.__class__
        if not cls._has_hpack(): return False
        if not cls._has_alpn(): return False
        if not cls._has_npn(): return False
        return True

    def _handle_exception(self, exception, connection):
        stream = exception.get_kwarg("stream")
        error_code = exception.get_kwarg("error_code", 0x00)
        message = exception.get_kwarg("message", "")
        ignore = exception.get_kwarg("ignore", False)
        self.warning(exception)
        self.log_stack()
        if ignore: return connection.send_ping(ack = True)
        if stream: return connection.error_stream(
            stream,
            error_code = error_code,
            message = message
        )
        return connection.error_connection(
            error_code = error_code,
            message = message
        )

    def _log_frame(self, connection, parser):
        self.debug(
            "Received frame 0x%02x (%s) for stream %d with length %d bytes" %\
            (parser.type, parser.type_s, parser.stream, parser.length)
        )

        self._log_frame_details(
            parser,
            parser.type_s,
            parser.flags,
            parser.payload,
            parser.stream,
            False
        )

    def _log_error(self, error_code, extra):
        message = netius.legacy.str(extra)
        self.warning(
            "Received error 0x%02x with message '%s'" %\
            (error_code, message)
        )

    def _log_send(self, connection, parser, type, flags, payload, stream):
        length = len(payload)
        type_s = parser.get_type_s(type)

        self.debug(
            "Sent frame 0x%02x (%s) for stream %d with length %d bytes" %\
            (type, type_s, stream, length)
        )

        self._log_frame_details(parser, type_s, flags, payload, stream, True)

    def _log_window(self, parser, stream, remote = False):
        name = "SEND" if remote else "RECV"
        connection = parser.connection
        window = connection.window if remote else connection.window_l
        self.debug("Connection %s window size is %d bytes" % (name, window))
        stream = parser._get_stream(stream, strict = False)
        if not stream: return
        window = stream.window if remote else stream.window_l
        self.debug(
            "Stream %d (dependency = %d, weight = %d) %s window size is %d bytes" %\
            (stream.identifier, stream.dependency, stream.weight, name, window)
        )

    def _log_frame_details(self, parser, type_s, flags, payload, stream, out):
        type_l = type_s.lower()
        method_s = "_log_frame_" + type_l
        if not hasattr(self, method_s): return
        method = getattr(self, method_s)
        method(parser, flags, payload, stream, out)

    def _log_frame_flags(self, type_s, *args):
        flags = ", ".join(args)
        pluralized = "flags" if len(args) > 1 else "flag"
        if flags: self.debug("%s with %s %s active" % (type_s, pluralized, flags))
        else: self.debug("Frame %s with no flags active" % type_s)

    def _log_frame_data(self, parser, flags, payload, stream, out):
        _stream = parser._get_stream(stream, strict = False)
        flags_l = self._flags_l(flags, (("END_STREAM", 0x01),))
        self._log_frame_flags("DATA", *flags_l)
        if _stream: self.debug("Frame DATA for path '%s'" % _stream.path_s)
        self._log_window(parser, stream, remote = out)

    def _log_frame_headers(self, parser, flags, payload, stream, out):
        flags_l = self._flags_l(
            flags,
            (
                ("END_STREAM", 0x01),
                ("END_HEADERS", 0x04),
                ("PADDED", 0x08),
                ("PRIORITY", 0x20)
            )
        )
        self._log_frame_flags("HEADERS", *flags_l)

    def _log_frame_rst_stream(self, parser, flags, payload, stream, out):
        error_code, = struct.unpack("!I", payload)
        self.debug("Frame RST_STREAM with error code %d" % error_code)

    def _log_frame_goaway(self, parser, flags, payload, stream, out):
        last_stream, error_code = struct.unpack("!II", payload[:8])
        extra = payload[8:]
        self.debug(
            "Frame GOAWAY with last stream %d, error code %d and message %s" %\
            (last_stream, error_code, extra)
        )

    def _log_frame_window_update(self, parser, flags, payload, stream, out):
        increment, = struct.unpack("!I", payload)
        self.debug("Frame WINDOW_UPDATE with increment %d" % increment)
        self._log_window(parser, stream, remote = not out)

    def _flags_l(self, flags, definition):
        flags_l = []
        for name, value in definition:
            valid = True if flags & value else False
            if not valid: continue
            flags_l.append(name)
        return flags_l



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import netius.common

class SSDPProtocol(netius.DatagramProtocol):
    """
    Protocol implementation of the SSDP protocol meant to be
    used under the UPnP standard for discovery and control of
    internet activated devices.

    Using this implementation it should be possible to discover
    and control devices like routes, media centers, etc.

    :see: http://en.wikipedia.org/wiki/Simple_Service_Discovery_Protocol
    """

    def on_data(self, address, data):
        netius.DatagramProtocol.on_data(self, address, data)
        self.parser = netius.common.HTTPParser(self, type = netius.common.RESPONSE)
        self.parser.bind("on_headers", self.on_headers_parser)
        self.parser.parse(data)
        self.parser.destroy()

    def on_headers_parser(self):
        headers = self.parser.get_headers()
        self.trigger("headers", self, self.parser, headers)

    def discover(self, target, *args, **kwargs):
        return self.method(
            "M-SEARCH",
            target,
            "ssdp:discover",
            *args,
            **kwargs
        )

    def method(
        self,
        method,
        target,
        namespace,
        mx = 3,
        path = "*",
        params = None,
        headers = None,
        data = None,
        host = "239.255.255.250",
        port = 1900,
        version = "HTTP/1.1",
        callback = None
    ):
        address = (host, port)

        headers = headers or dict()
        headers["ST"] = target
        headers["Man"] = "\"" + namespace + "\""
        headers["MX"] = str(mx)
        headers["Host"] = "%s:%d" % address

        buffer = []
        buffer.append("%s %s %s\r\n" % (method, path, version))
        for key, value in netius.legacy.iteritems(headers):
            key = netius.common.header_up(key)
            if not isinstance(value, list): value = (value,)
            for _value in value: buffer.append("%s: %s\r\n" % (key, _value))
        buffer.append("\r\n")
        buffer_data = "".join(buffer)

        self.send(buffer_data, address)
        data and self.send(data, address, callback = callback)

class SSDPClient(netius.ClientAgent):

    protocol = SSDPProtocol

    @classmethod
    def discover_s(cls, target, *args, **kwargs):
        return cls.method_s(
            "M-SEARCH",
            target,
            "ssdp:discover",
            *args,
            **kwargs
        )

    @classmethod
    def method_s(
        cls,
        method,
        target,
        namespace,
        mx = 3,
        path = "*",
        params = None,
        headers = None,
        data = None,
        host = "239.255.255.250",
        port = 1900,
        version = "HTTP/1.1",
        callback = None,
        loop = None
    ):
        address = (host, port)
        protocol = cls.protocol()

        def on_connect(result):
            _transport, protocol = result
            protocol.method(
                method,
                target,
                namespace,
                mx = mx,
                path = path,
                params = params,
                headers = headers,
                data = data,
                host = host,
                port = port,
                version = version,
                callback = callback
            )

        loop = netius.build_datagram(
            lambda: protocol,
            callback = on_connect,
            loop = loop,
            remote_addr = address
        )

        return loop, protocol

if __name__ == "__main__":
    def on_headers(client, parser, headers):
        print(headers)

        # retrieves the currently associated loop using
        # netius base infra-structure and then runs the
        # stop operation on the next tick end
        netius.compat_loop(loop).stop()

    target = netius.conf("SSDP_TARGET", "urn:schemas-upnp-org:device:InternetGatewayDevice:1")

    loop, protocol = SSDPClient.discover_s(target)

    protocol.bind("headers", on_headers)

    loop.run_forever()
    loop.close()
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import netius

from . import parser

class POPParser(parser.Parser):

    def __init__(self, owner, store = False):
        parser.Parser.__init__(self, owner)

        self.buffer = []

    def parse(self, data):
        """
        Parses the provided data chunk, changing the current
        state of the parser accordingly and returning the
        number of processed bytes from it.

        :type data: String
        :param data: The string containing the data to be parsed
        in the current parse operation.
        :rtype: int
        :return: The amount of bytes of the data string that have
        been "parsed" in the current parse operation.
        """

        parser.Parser.parse(self, data)

        # retrieves the size of the data that has been sent for parsing
        # and saves it under the size original variable
        size = len(data)
        size_o = size

        # iterates continuously to try to process all that
        # data that has been sent for processing
        while size > 0:
            # retrieves the parsing method for the current
            # state and then runs it retrieving the number
            # of valid parsed bytes in case this value is
            # zero the parsing iteration is broken
            method = self._parse_line
            count = method(data)
            if count == 0: break

            # decrements the size of the data buffer by the
            # size of the parsed bytes and then retrieves the
            # sub part of the data buffer as the new data buffer
            size -= count
            data = data[count:]

        # in case not all of the data has been processed
        # must add it to the buffer so that it may be used
        # latter in the next parsing of the message
        if size > 0: self.buffer.append(data)

        # returns the number of read (processed) bytes of the
        # data that has been sent to the parser
        return size_o - size

    def _parse_line(self, data):
        # tries to find the new line character in the currently received
        # data in case it's not found returns immediately with no data processed
        index = data.find(b"\n")
        if index == -1: return 0

        # adds the partial data (until new line) to the current buffer and
        # then joins it retrieving the current line, then deletes the buffer
        # as it's not longer going to be used
        self.buffer.append(data[:index])
        self.line_s = b"".join(self.buffer).rstrip()
        self.line_s = netius.legacy.str(self.line_s)
        del self.buffer[:]

        # splits the provided line into the code and message parts in case
        # the split is not successful (not enough information) then an extra
        # value is added to the sequence of values for compatibility
        values = self.line_s.split(" ", 1)
        if not len(values) > 1: values.append(b"")

        # unpacks the set of values that have just been parsed into the code
        # and the message items as expected by the pop specification
        code, message = values

        # triggers the on line event so that the listeners are notified
        # about the end of the parsing of the pop line and then
        # returns the count of the parsed bytes of the message
        self.trigger("on_line", code, message)
        return index + 1



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from . import server

from .common import * #@UnusedWildImport

class Container(Base):

    def __init__(self, *args, **kwargs):
        Base.__init__(self, *args, **kwargs)
        self.owner = None
        self.bases = []

    def start(self, owner):
        # sets the current polling structure of the owner in the container
        # it's important to use the already initialized poll in the container
        # so that the requested environment (host, port , etc.) is used note
        # that event the same logger is used for the container (logic propagation)
        self.owner = owner
        self.poll_c = owner.poll_c
        self.poll = owner.poll
        self.poll_name = owner.poll_name
        self.poll_timeout = owner.poll_timeout
        self.level = owner.level
        self.logger = owner.logger
        self._loaded = True

        # sets the owner of the current container as the owner of the poll
        # this avoids problems with the cleanup operation for the poll
        owner.poll_owner = True

        # runs the starting operation in the complete set of base structures
        # registered under the container, this is the required operation in
        # order to propagate the changed in the container to the bases
        self.start_all()

        # calls the super method of the base for the current container this should
        # start the event loop for the container (blocking call)
        Base.start(self)

    def cleanup(self):
        Base.cleanup(self)

        # unsets the owner of the container, this should diminish the chance of
        # memory leaks due to cycle references (required to avoid problems)
        self.owner = None

        # iterates over all the bases registered and propagates the cleanup operation
        # over them, deleting the list of bases afterwards (no more usage for them)
        for base in self.bases: base.cleanup()
        del self.bases[:]

        # unbinds the start operation from the on start event, as this is no longer
        # required, should re-register for it on "next start" event
        self.unbind("start", self.on_start)

    def loop(self):
        # iterates continuously while the running flag
        # is set, once it becomes unset the loop breaks
        # at the next execution cycle
        while self._running:
            # calls the base tick int handler indicating that a new
            # tick loop iteration is going to be started, all the
            # "in between loop" operation should be performed in this
            # callback as this is the "space" they have for execution
            self.ticks()

            # updates the current state to poll to indicate
            # that the base service is selecting the connections
            self.set_state(STATE_POLL)

            # runs the "owner" based version of the poll operation
            # so that the poll results are indexed by their owner
            # reference to be easily routed to the base services
            result = self.poll.poll_owner()
            for base, values in legacy.iteritems(result):
                reads, writes, errors = values
                base.reads(reads)
                base.writes(writes)
                base.errors(errors)

    def ticks(self):
        self.set_state(STATE_TICK)
        self._lid = (self._lid + 1) % 2147483647
        for base in self.bases: base.ticks()

    def connections_dict(self, full = False):
        all = dict()
        for base in self.bases:
            is_owner = base == self.owner
            if is_owner: connections = base.connections_dict(
                full = full, parent = True
            )
            else: connections = base.connections_dict(full = full)
            all[base.name] = connections
        return all

    def connection_dict(self, id, full = False):
        connection = None
        for base in self.bases:
            for _connection in base.connections:
                if not _connection.id == id: continue
                connection = _connection
                break
            if connection: break
        if not connection: return None
        return connection.info_dict(full = full)

    def on_start(self):
        Base.on_start(self)
        self.apply_all()
        self.call_all("on_start")

    def on_stop(self):
        Base.on_stop(self)
        self.call_all("on_stop")

    def add_base(self, base):
        self.apply_base(base)
        self.bases.append(base)

    def remove_base(self, base):
        self.bases.remove(base)

    def start_base(self, base):
        base.level = self.level
        base.logger = self.logger
        base.load()

    def start_all(self):
        for base in self.bases: self.start_base(base)

    def apply_all(self):
        for base in self.bases: self.apply_base(base)

    def apply_base(self, base):
        base.tid = self.tid
        base.poll = self.poll
        base.level = self.level
        base.logger = self.logger
        base.poll_owner = base == self.owner

    def call_all(self, name, *args, **kwargs):
        for base in self.bases:
            method = getattr(base, name)
            method(*args, **kwargs)

    def trigger_all(self, name, *args, **kwargs):
        for base in self.bases: base.trigger(name, base, *args, **kwargs)

class ContainerServer(server.StreamServer):

    def __init__(self, *args, **kwargs):
        server.StreamServer.__init__(self, *args, **kwargs)
        self.container = Container(*args, **kwargs)
        self.add_base(self)

    def start(self):
        # starts the container this should trigger the start of the
        # event loop in the container and the proper listening of all
        # the connections in the current environment
        self.container.start(self)

    def stop(self):
        # verifies if there's a container object currently defined in
        # the object and in case it does exist propagates the stop call
        # to the container so that the proper stop operation is performed
        if not self.container: return
        self.container.stop()

    def cleanup(self):
        server.StreamServer.cleanup(self)

        # saves the container reference so that it may be used latter
        # and then unsets it under the current instance
        container = self.container
        self.container = None

        # verifies if the container is valid and if that's not the case
        # returns the control flow immediately (as expected)
        if not container: return

        # runs the cleanup operation on the cleanup, this should properly
        # propagate the operation to the owner container (as expected)
        container.cleanup()

    def add_base(self, base):
        self.container.add_base(base)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import uuid
import hashlib
import datetime

import netius.common
import netius.clients
import netius.servers

class RelaySMTPServer(netius.servers.SMTPServer):
    """
    Relay version of the smtp server that relays messages
    that are not considered to be local to other servers.

    The servers uses the default smtp client implementation
    to relay the messages.
    """

    def __init__(self, *args, **kwargs):
        netius.servers.SMTPServer.__init__(self, *args, **kwargs)
        self.dkim = {}

    def on_serve(self):
        netius.servers.SMTPServer.on_serve(self)
        self.dkim = self.get_env("DKIM", self.dkim)
        dkim_l = len(self.dkim)
        self.info("Starting Relay SMTP server with %d DKIM registers  ..." % dkim_l)

    def on_header_smtp(self, connection, from_l, to_l):
        netius.servers.SMTPServer.on_header_smtp(self, connection, from_l, to_l)

        # retrieves the list or remote emails for each a relay
        # operation will have to be performed as requested by
        # the current smtp specification (federation based)
        remotes = self._remotes(to_l)

        # updates the current connection with the list of remote
        # emails that have to be relayed and starts the buffer
        # that will hold the complete message for relay
        connection.remotes = remotes
        connection.relay = []

    def on_data_smtp(self, connection, data):
        netius.servers.SMTPServer.on_data_smtp(self, connection, data)

        # verifies if there're remote addresses in the current
        # connection's message and if there is adds the received
        # data to the current relay buffer that is used
        if not connection.remotes: return
        connection.relay.append(data)

    def on_message_smtp(self, connection):
        netius.servers.SMTPServer.on_message_smtp(self, connection)

        # in case there's no remotes list in the current connection
        # there's no need to proceed as no relay is required
        if not connection.remotes: return

        # joins the current relay buffer to create the full message
        # data and then removes the (non required) termination value
        # from it to avoid any possible problems with extra size
        data_s = b"".join(connection.relay)
        data_s = data_s[:netius.servers.TERMINATION_SIZE * -1]

        # retrieves the list of "froms" for the connection and then
        # sends the message for relay to all of the current remotes
        froms = self._emails(connection.from_l, prefix = "from")
        self.relay(connection, froms, connection.remotes, data_s)

    def relay(self, connection, froms, tos, contents):
        # verifies that the current connection has an authenticated user
        # and if not raises an exception as the authentication is mandatory
        # for the relaying of message under the "default" policy
        if not hasattr(connection, "username") or not connection.username:
            raise netius.SecurityError("User is not authenticated")

        # retrieves the current date value formatted according to
        # the smtp based specification string value, this value
        # is going to be used for the replacement of the header
        date = self.date()

        # retrieves the first email from the froms list as this is
        # the one that is going to be used for message id generation
        # and then generates a new "temporary" message id
        first = froms[0]
        message_id = self.message_id(connection, email = first)

        # parses the provided contents as mime text and then appends
        # the various extra fields so that the relay operation is
        # considered valid and then re-joins the message as a string
        headers, body = netius.common.rfc822_parse(contents)
        received = connection.received_s()
        message_id = headers.pop("Message-Id", message_id)
        message_id = headers.pop("Message-ID", message_id)
        headers.set("Date", date)
        headers.set("Received", received)
        headers.set("Message-ID", message_id)
        contents = netius.common.rfc822_join(headers, body)

        # tries to sign the message using dkim, the server is going to
        # search the current registry, trying to find a registry for the
        # domain of the sender and if it finds one signs the message using
        # the information provided by the registry
        contents = self.dkim_contents(contents, email = first)

        # creates the callback that will close the client once the message
        # is sent to all the recipients (better auto close support), note
        # that multiple smtp session may be created for the message so that
        # all the hosts associated with the recipients are notified
        callback = lambda smtp_client: smtp_client.close()

        # generates a new smtp client for the sending of the message,
        # uses the current host for identification and then triggers
        # the message event to send the message to the target host
        smtp_client = netius.clients.SMTPClient(host = self.host)
        smtp_client.message(
            froms,
            tos,
            contents,
            mark = False,
            callback = callback
        )

    def date(self):
        date_time = datetime.datetime.utcnow()
        return date_time.strftime("%a, %d %b %Y %H:%M:%S +0000")

    def message_id(self, connection, email = "user@localhost"):
        _user, domain = email.split("@", 1)
        domain = self.host or domain
        identifier = str(uuid.uuid4())
        identifier = netius.legacy.bytes(identifier)
        identifier = hashlib.sha1(identifier).hexdigest()
        identifier = identifier.upper()
        identifier = connection.identifier or identifier
        return "<%s@%s>" % (identifier, domain)

    def dkim_contents(self, contents, email = "user@localhost", creation = None):
        _user, domain = email.split("@", 1)
        register = self.dkim.get(domain, None)
        if not register: return contents

        key_path = register.get("key", None)
        key_b64 = register.get("key_b64", None)
        selector = register["selector"]
        domain = register["domain"]

        contents = contents.lstrip()

        if key_path: private_key = netius.common.open_private_key(key_path)
        elif key_b64: private_key = netius.common.open_private_key_b64(key_b64)
        else: raise netius.SecurityError("No private key provided")

        signature = netius.common.dkim_sign(
            contents,
            selector,
            domain,
            private_key,
            identity = email,
            creation = creation
        )

        return signature + contents

if __name__ == "__main__":
    import logging
    server = RelaySMTPServer(level = logging.DEBUG)
    server.serve(env = True)
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import ctypes
import socket
import threading

import netius

CALLABLE_WORK = 1

class Thread(threading.Thread):

    def __init__(self, identifier, owner = None, *args, **kwargs):
        threading.Thread.__init__(self, *args, **kwargs)
        self.identifier = identifier
        self.owner = owner
        self._run = False

    @classmethod
    def available(self):
        return True

    def stop(self):
        self._run = False

    def run(self):
        threading.Thread.run(self)
        self._run = True
        while self._run: self.tick()

    def tick(self):
        self.owner.condition.acquire()
        while not self.owner.peek() and self._run:
            self.owner.condition.wait()
        try:
            if not self._run: return
            work = self.owner.pop()
        finally:
            self.owner.condition.release()
        self.execute(work)

    def execute(self, work):
        type = work[0]
        if type == CALLABLE_WORK: work[1]()
        else: raise netius.NotImplemented("Cannot execute type '%d'" % type)

class ThreadPool(object):

    def __init__(self, base = Thread, count = 32):
        self.base = base
        self.count = count
        self.instances = []
        self.queue = []
        self.condition = threading.Condition()
        self._built = False

    def start(self):
        self.build()
        for instance in self.instances:
            instance.start()

    def stop(self, join = True):
        for instance in self.instances: instance.stop()
        self.condition.acquire()
        try: self.condition.notify_all()
        finally: self.condition.release()
        if not join: return
        for instance in self.instances: instance.join()

    def build(self):
        if self._built: return
        for index in range(self.count):
            instance = self.base(index, owner = self)
            self.instances.append(instance)
        self._built = True

    def peek(self):
        if not self.queue: return None
        return self.queue[0]

    def pop(self, lock = True):
        lock and self.condition.acquire()
        try: value = self.queue.pop(0)
        finally: lock and self.condition.release()
        return value

    def push(self, work, lock = True):
        lock and self.condition.acquire()
        try:
            value = self.queue.append(work)
            self.condition.notify()
        finally:
            lock and self.condition.release()
        return value

    def push_callable(self, callable):
        work = (CALLABLE_WORK, callable)
        self.push(work)

class EventPool(ThreadPool):

    def __init__(self, base = Thread, count = 32):
        ThreadPool.__init__(self, base = base, count = count)
        self.events = []
        self.event_lock = threading.RLock()
        self._eventfd = None

    def stop(self, join = True):
        ThreadPool.stop(self, join = join)
        if not self._eventfd: return
        self._eventfd.close()
        self._eventfd = None

    def push_event(self, event):
        self.event_lock.acquire()
        try: self.events.append(event)
        finally: self.event_lock.release()
        self.notify()

    def pop_event(self):
        self.event_lock.acquire()
        try: event = self.events.pop(0)
        finally: self.event_lock.release()
        return event

    def pop_all(self, denotify = False):
        self.event_lock.acquire()
        try:
            events = list(self.events)
            del self.events[:]
            if events and denotify: self.denotify()
        finally:
            self.event_lock.release()
        return events

    def notify(self):
        if not self._eventfd: return
        self._eventfd.notify()

    def denotify(self):
        if not self._eventfd: return
        self._eventfd.denotify()

    def eventfd(self):
        if self._eventfd: return self._eventfd
        if UnixEventFile.available():
            self._eventfd = UnixEventFile()
        elif PipeEventFile.available():
            self._eventfd = PipeEventFile()
        else:
            self._eventfd = SocketEventFile()
        return self._eventfd

class EventFile(object):

    def __init__(self, *args, **kwargs):
        self._rfileno = None
        self._wfileno = None
        self.closed = False

    def close(self):
        self.closed = True

    def fileno(self):
        return self.rfileno()

    def rfileno(self):
        return self._rfileno

    def wfileno(self):
        return self._wfileno

    def notify(self):
        raise netius.NotImplemented("Missing implementation")

    def denotify(self):
        raise netius.NotImplemented("Missing implementation")

class UnixEventFile(EventFile):

    _LIBC = None

    def __init__(self, *args, **kwargs):
        EventFile.__init__(self, *args, **kwargs)
        cls = self.__class__
        init_val = kwargs.get("init_val", 0)
        flags = kwargs.get("flags", 0)
        libc = cls.libc()
        self._rfileno = libc.eventfd(init_val, flags)
        self._wfileno = self._rfileno

    @classmethod
    def available(cls):
        if not os.name == "posix": return False
        return True if cls.libc() else False

    @classmethod
    def libc(cls):
        if cls._LIBC: return cls._LIBC
        try: cls._LIBC = ctypes.cdll.LoadLibrary("libc.so.6")
        except Exception: return None
        return cls._LIBC

    def close(self):
        EventFile.close(self)
        os.close(self._wfileno)

    def notify(self):
        self._write(1)

    def denotify(self):
        self._read()

    def _read(self, length = 8):
        if self.closed: return None
        return os.read(self._rfileno, length)

    def _write(self, value):
        if self.closed: return
        os.write(self._wfileno, ctypes.c_ulonglong(value))

class PipeEventFile(EventFile):

    def __init__(self, *args, **kwargs):
        import fcntl
        EventFile.__init__(self, *args, **kwargs)
        self._rfileno, self._wfileno = os.pipe()
        fcntl.fcntl(self._rfileno, fcntl.F_SETFL, os.O_NONBLOCK) #@UndefinedVariable
        fcntl.fcntl(self._wfileno, fcntl.F_SETFL, os.O_NONBLOCK) #@UndefinedVariable
        self._read_file = os.fdopen(self._rfileno, "rb", 0)
        self._write_file = os.fdopen(self._wfileno, "wb", 0)
        self._lock = threading.RLock()
        self._count = 0

    @classmethod
    def available(cls):
        if not os.name == "posix": return False
        if not hasattr(os, "pipe"): return False
        return True

    def close(self):
        EventFile.close(self)
        self._read_file.close()
        self._write_file.close()

    def notify(self):
        self._lock.acquire()
        try:
            self._write(b"1")
            self._count += 1
        finally:
            self._lock.release()

    def denotify(self):
        self._lock.acquire()
        try:
            data = self._read()
            self._count -= len(data)
        finally:
            self._lock.release()

    def _read(self, length = 4096):
        if self.closed: return None
        return self._read_file.read(length)

    def _write(self, data):
        if self.closed: return
        self._write_file.write(data)

class SocketEventFile(EventFile):

    def __init__(self, *args, **kwargs):
        EventFile.__init__(self, *args, **kwargs)
        temp_socket = socket.socket()
        temp_socket.bind(("127.0.0.1", 0))
        temp_socket.listen(1)
        hostname = temp_socket.getsockname()
        self._read_socket = socket.create_connection(hostname)
        self._write_socket, _port = temp_socket.accept()
        self._rfileno = self._read_socket.fileno()
        self._wfileno = self._write_socket.fileno()
        self._lock = threading.RLock()
        self._count = 0
        temp_socket.close()

    def close(self):
        EventFile.close(self)
        self._read_socket.close()
        self._write_socket.close()

    def notify(self):
        self._lock.acquire()
        try:
            self._write(b"1")
            self._count += 1
        finally:
            self._lock.release()

    def denotify(self):
        self._lock.acquire()
        try:
            if self._count == 0: return
            data = self._read()
            self._count -= len(data)
        finally:
            self._lock.release()

    def _read(self, length = 4096):
        if self.closed: return None
        return self._read_socket.recv(length)

    def _write(self, data):
        if self.closed: return
        self._write_socket.send(data)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import sys
import copy
import math
import uuid
import time
import struct
import hashlib

import netius.common
import netius.clients

REFRESH_TIME = 30.0
""" The default time in between refresh call in the torrent task
each refresh operation should perform some operations (eg: DHT
refresh, tracker re-retrieval, etc) """

ID_STRING = "NE1000"
""" Text value that is going to be used to identify the agent
of torrent against the other peers, should be a join of both
the abbreviation of the agent name and version """

BLOCK_SIZE = 16384
""" The typical size of block that is going to be retrieved
using the current torrent infra-structure, this value conditions
most of the torrent operations and should be defined carefully """

THRESHOLD_END = 10485760
""" The threshold value from which the task is considered to be
under the ending stage, from this stage on a new strategy for the
download may apply as it is more difficult to get blocks """

MAX_MISSING = 16
""" The maximum number of unmarked values to be displayed in missing,
this controls the size of the missing lists, note that this is only
a default value, that may be overriden at runtime """

PEER_PATHS = ("peers.txt", "~/peers.txt", "\\peers.txt")
""" The sequence defining the various paths that are going to be search
trying to find the (static) peers file with format host:ip in each line """

class Pieces(netius.Observable):
    """
    Class that represents the logical structure of a file that is
    divided into pieces and blocks as a hierarchy, this class is
    responsible for the management of the data structures of such
    data storage model.

    A piece is considered to be the basic unit of a torrent file
    and each piece is composed by multiple blocks, note that the
    number of blocks is fixed for all the pieces of a file.
    """

    def __init__(self, length, number_pieces, number_blocks):
        netius.Observable.__init__(self)

        self.length = length
        self.piece_length = number_blocks * BLOCK_SIZE
        self.number_pieces = number_pieces
        self.number_blocks = number_blocks
        self.final_blocks = self.piece_blocks(self.number_pieces - 1)
        self.bitfield = [True for _index in netius.legacy.xrange(number_pieces)]
        self.mask = [True for _index in netius.legacy.xrange(self.total_blocks)]

    def piece(self, index):
        return self.bitfield[index]

    def piece_blocks(self, index):
        is_last = index == self.number_pieces - 1
        if not is_last: return self.number_blocks
        piece_size = self.piece_size(index)
        number_blocks = math.ceil(piece_size / float(BLOCK_SIZE))
        return int(number_blocks)

    def piece_size(self, index):
        is_last = index == self.number_pieces - 1
        if not is_last: return self.number_blocks * BLOCK_SIZE
        modulus = self.length % self.piece_length
        if modulus == 0: return self.piece_length
        return modulus

    def block(self, index, begin):
        base = index * self.number_blocks
        block_index = begin // BLOCK_SIZE
        return self.mask[base + block_index]

    def block_size(self, index, begin):
        block_index = begin // BLOCK_SIZE
        is_last_piece = index == self.number_pieces - 1
        if not is_last_piece: return BLOCK_SIZE
        is_last_block = block_index == self.final_blocks - 1
        if not is_last_block: return BLOCK_SIZE
        piece_size = self.piece_size(index)
        modulus = piece_size % BLOCK_SIZE
        if modulus == 0: return BLOCK_SIZE
        return modulus

    def pop_block(self, bitfield, mark = True):
        index = 0
        result = self._and(bitfield, self.bitfield)
        for bit in result:
            if bit == True: break
            index += 1

        if index == len(result): return None

        begin = self.update_block(index, mark = mark)
        length = self.block_size(index, begin)
        return (index, begin, length)

    def push_block(self, index, begin):
        self.mark_block(index, begin, value = True)

    def mark_piece(self, index, value = False):
        base = index * self.number_blocks
        block_count = self.piece_blocks(index)

        for block_index in netius.legacy.xrange(block_count):
            self.mask[base + block_index] = value

        self.bitfield[index] = value

    def mark_block(self, index, begin, value = False):
        base = index * self.number_blocks
        block_index = begin // BLOCK_SIZE
        self.mask[base + block_index] = value
        self.trigger("block", self, index, begin)
        self.update_piece(index)

    def update_block(self, index, mark = True):
        base = index * self.number_blocks
        block_count = self.piece_blocks(index)

        for block_index in netius.legacy.xrange(block_count):
            state = self.mask[base + block_index]
            if state == True: break

        begin = block_index * BLOCK_SIZE
        if mark: self.mark_block(index, begin)
        return begin

    def update_piece(self, index):
        # calculates the base index value for the block sequence
        # of the current piece (going to be used in access), then
        # determines the total number of block for piece to update
        # and then sets the initial piece state as false (not marked)
        base = index * self.number_blocks
        block_count = self.piece_blocks(index)
        piece_state = False

        # iterates over the complete set of blocks for the current
        # piece trying to determine if it has already been completely
        # unmarked (all the blocks unmarked accordingly)
        for block_index in netius.legacy.xrange(block_count):
            state = self.mask[base + block_index]
            if state == False: continue
            piece_state = True
            break

        # updates the state of the current piece in the bit field,
        # note that the false value indicates that the piece has been
        # unmarked (and this is considered the objective)
        self.bitfield[index] = piece_state
        if piece_state == True: return

        # triggers the piece event indicating that a new piece has
        # been completely unmarked according to rules
        self.trigger("piece", self, index)

        # iterates over the complete set of bit values in the (pieces)
        # bit field to verify if the file has been completely unmarked
        # in case it did not returns the control flow to caller
        for bit in self.bitfield:
            if bit == True: return

        # triggers the complete event to any of the handlers indicating
        # that the current torrent file has been completely unmarked
        # and then no more pieces are pending to be unmarked
        self.trigger("complete", self)

    @property
    def total_pieces(self):
        return self.number_pieces

    @property
    def marked_pieces(self):
        counter = 0
        for bit in self.bitfield:
            if bit == True: continue
            counter += 1
        return counter

    @property
    def missing_pieces(self, max_missing = MAX_MISSING):
        missing_count = self.total_pieces - self.marked_pieces
        if missing_count > max_missing: return []
        missing = []
        for index in netius.legacy.xrange(self.total_pieces):
            bit = self.bitfield[index]
            if bit == False: continue
            missing.append(index)
        return missing

    @property
    def total_blocks(self):
        base_blocks = (self.number_pieces - 1) * self.number_blocks
        return base_blocks + self.final_blocks

    @property
    def marked_blocks(self):
        counter = 0
        for bit in self.mask:
            if bit == True: continue
            counter += 1
        return counter

    @property
    def missing_blocks(self, max_missing = MAX_MISSING):
        missing_count = self.total_blocks - self.marked_blocks
        if missing_count > max_missing: return []
        missing = []
        for index in netius.legacy.xrange(self.total_blocks):
            bit = self.mask[index]
            if bit == False: continue
            missing.append(index)
        return missing

    def _and(self, first, second):
        result = []
        for _first, _second in zip(first, second):
            if _first and _second: value = True
            else: value = False
            result.append(value)
        return result

class TorrentTask(netius.Observable):
    """
    Describes a task (operation) that is going to be performed
    using the peer to peer mesh network of the torrent protocol.

    Each of the download operations should be able to be described
    by this task object (for latter reference).

    This should be considered the main interface to interact from
    a developer point of view, as such the methods should represent
    a proper easily described interface.
    """

    def __init__(self, owner, target_path, torrent_path = None, info_hash = None):
        netius.Observable.__init__(self)

        self.owner = owner
        self.target_path = target_path
        self.torrent_path = torrent_path
        self.info_hash = info_hash
        self.start = time.time()
        self.uploaded = 0
        self.downloaded = 0
        self.unchoked = 0
        self.next_refresh = self.start + REFRESH_TIME
        self.connections = []
        self.peers = []
        self.peers_m = {}

    def load(self):
        if self.torrent_path: self.info = self.load_info(self.torrent_path)
        else: self.info = dict(info_hash = self.info_hash)

        self.pieces_tracker()
        self.peers_dht()
        self.peers_tracker()
        self.peers_file()

        self.load_file()
        self.load_pieces()

    def unload(self):
        self.owner = None
        self.unload_file()
        self.unload_pieces()
        self.disconnect_peers()

    def on_close(self, connection):
        is_unchoked = connection.choked == netius.clients.UNCHOKED
        self.connections.remove(connection)
        self.unchoked -= 1 if is_unchoked else 0

    def ticks(self):
        if time.time() < self.next_refresh: return
        self.refresh()

    def refresh(self):
        self.peers_dht()
        self.peers_tracker()
        self.connect_peers()
        self.next_refresh = time.time() + REFRESH_TIME
        self.trigger("refresh", self)

    def on_choked(self, connection):
        self.unchoked -= 1

    def on_unchoked(self, connection):
        self.unchoked += 1

    def on_block(self, pieces, index, begin):
        self.trigger("block", self, index, begin)

    def on_piece(self, pieces, index):
        try: self.verify_piece(index)
        except netius.DataError:
            self.refute_piece(index)
        else:
            self.confirm_piece(index)
            self.trigger("piece", self, index)

    def on_complete(self, pieces):
        self.trigger("complete", self)

    def on_dht(self, response):
        # verifies if the response is valid and in case it's not
        # returns immediately to avoid any erroneous parsing
        if not response: return

        # retrieves the payload for the response and then uses it
        # to retrieves the nodes part of the response for parsing
        # of the peers that are going to be added (to the task)
        payload = response.get_payload()
        nodes = payload.get("nodes", "")

        # creates the list that will hold the final set of peers
        # parsed from the nodes string, this is going to be used
        # to extend the list of peers in the task
        peers = []

        # splits the current nodes list into a set of chunks of
        # a pre-defined size and then iterates over all of them
        # creating the proper peer dictionary for each of them
        chunks = [chunk for chunk in netius.common.chunks(nodes, 26)]
        for chunk in chunks:
            chunk = netius.legacy.bytes(chunk)
            peer_id, address, port = struct.unpack("!20sLH", chunk)
            ip = netius.common.addr_to_ip4(address)
            peer = dict(id = peer_id, ip = ip, port = port)
            peers.append(peer)

        # in case no valid peers have been parsed there's no need
        # to continue with the processing, nothing to be done
        if not peers: return

        # extends the currently defined peers list in the current
        # torrent task with the ones that have been discovered
        self.extend_peers(peers)

        # retrieves the reference to the host id from the request
        # that originated the current response and then converts it
        # into the proper string representation to be used in logging
        request = response.request
        host = request.host

        # prints a debug message about the peer loading that has just occurred, this
        # may be used for the purpose of development (and traceability)
        self.owner.debug("Received %d peers from DHT peer '%s'" % (len(peers), host))

    def on_tracker(self, client, parser, result):
        # extracts the data (string) contents of the http response and in case
        # there're none of them continues the loop as there's nothing to be
        # processed from this tracker response (invalid response)
        data = result["data"]
        if not data: return

        # tries to decode the provided data from the tracker using the bencoder
        # and extracts the peers part of the message to be processed
        response = netius.common.bdecode(data)
        peers = response["peers"]

        # verifies if the provided peers part is not compact (already a dictionary)
        # if that's the case there's nothing remaining to be done, otherwise extra
        # processing must be done to
        if isinstance(peers, dict): self.extend_peers(peers)

        # need to normalize the peer structure by decoding the peers string into a
        # set of address port sub strings (as defined in torrent specification)
        else:
            peers = [peer for peer in netius.common.chunks(peers, 6)]
            for peer in peers:
                peer = netius.legacy.bytes(peer)
                address, port = struct.unpack("!LH", peer)
                ip = netius.common.addr_to_ip4(address)
                peer = dict(ip = ip, port = port)
                self.add_peer(peer)

        # prints a debug message about the peer loading that has just occurred, this
        # may be used for the purpose of development (and traceability)
        self.owner.debug("Received %d peers from '%s'" % (len(peers), parser.owner.base))

        # refreshes the connection with the peers because new peers have been added
        # to the current task and there may be new connections pending
        self.connect_peers()

    def load_info(self, torrent_path):
        file = open(torrent_path, "rb")
        try: data = file.read()
        finally: file.close()

        struct = netius.common.bdecode(data)
        struct["info_hash"] = self.info_hash = netius.common.info_hash(struct)
        return struct

    def load_file(self):
        if self._is_single(): return self.load_single()
        else: return self.load_multiple()

    def load_single(self):
        # retrieves the length of the current (single file) and
        # the name of the associated file
        size = self.info["length"]
        name = self.info["info"]["name"]

        # runs the normalization process on the target path so that
        # it may be used on a more flexible way
        target_path = os.path.expanduser(self.target_path)
        target_path = os.path.normpath(target_path)

        # determines if the target path is a directory and if that's
        # not the case creates the appropriate directories so that
        # they area available for the file stream creation
        is_dir = os.path.isdir(target_path)
        if not is_dir: os.makedirs(target_path)

        # creates the "final" file path from the target path and the
        # name of the file and then constructs a file stream with the
        # path and the size information and opens it, note that the
        # opening operation is expensive as it allocates the file
        file_path = os.path.join(target_path, name)
        self.file = netius.common.FileStream(file_path, size)
        self.file.open()

    def load_multiple(self):
        files = self.info["files"]
        size = self.info["length"]
        name = self.info["info"]["name"]

        target_path = os.path.expanduser(self.target_path)
        target_path = os.path.normpath(target_path)

        dir_path = os.path.join(target_path, name)
        is_dir = os.path.isdir(dir_path)
        if not is_dir: os.makedirs(dir_path)

        self.file = netius.common.FilesStream(dir_path, size, files)
        self.file.open()

    def unload_file(self):
        if not self.file: return
        self.file.close()
        self.file = None

    def load_pieces(self):
        length = self.info["length"]
        number_pieces = self.info["number_pieces"]
        number_blocks = self.info["number_blocks"]
        self.requested = Pieces(length, number_pieces, number_blocks)
        self.stored = Pieces(length, number_pieces, number_blocks)
        self.stored.bind("block", self.on_block)
        self.stored.bind("piece", self.on_piece)
        self.stored.bind("complete", self.on_complete)

    def unload_pieces(self):
        if self.requested: self.requested.destroy()
        if self.stored: self.stored.destroy()
        self.requested = None
        self.stored = None

    def pieces_tracker(self):
        info = self.info.get("info", {})
        pieces = info.get("pieces", "")
        length = info.get("length", None)
        files = info.get("files", [])
        piece_length = info.get("piece length", 1)
        number_blocks = math.ceil(float(piece_length) / float(BLOCK_SIZE))
        number_blocks = int(number_blocks)
        pieces_l = [piece for piece in netius.common.chunks(pieces, 20)]
        pieces_count = len(pieces_l)
        files_length = sum(file["length"] for file in files)
        self.info["pieces"] = pieces_l
        self.info["length"] = length or files_length or pieces_count * piece_length
        self.info["files"] = files
        self.info["number_pieces"] = pieces_count
        self.info["number_blocks"] = number_blocks

    def set_data(self, data, index, begin):
        # retrieves the current status of the block in the stored
        # pieces structure and in case it's already stored returns
        # immediately as this is a duplicated block setting, possible
        # in the last part of the file retrieval (end game)
        block = self.stored.block(index, begin)
        if not block: return

        # retrieves the size of a piece and uses that value together
        # with the block begin offset to calculate the final file offset
        # value to be passed to the write data operations (for handling)
        piece_length = self.info["info"]["piece length"]
        offset = index * piece_length + begin
        self.write_data(data, offset)

        # marks the current block as stored so that no other equivalent
        # operation is performed (avoiding duplicated operations)
        self.stored.mark_block(index, begin)

    def write_data(self, data, offset):
        # seek the proper file position (according to passed offset)
        # and then writes the received data under that position,
        # flushing the file contents afterwards to avoid file corruption
        self.file.seek(offset)
        self.file.write(data)
        self.file.flush()

    def set_dht(self, peer_t, port):
        # tries to retrieve the peer associated with the provided peer tuple
        # and in case it succeeds sets the proper DHT (port) value in the peer
        # so that it may latter be used for DHT based operations
        peer = self.peers_m.get(peer_t, None)
        if not peer: return
        peer["dht"] = port

    def peers_dht(self):
        if not self.info_hash: return
        for peer in self.peers:
            port = peer.get("dht", None)
            if not port: continue
            host = peer["ip"]
            self.owner.dht_client.get_peers(
                host = host,
                port = port,
                peer_id = self.owner.peer_id,
                info_hash = self.info_hash,
                callback = self.on_dht
            )
            self.owner.debug("Requested peers from DHT peer '%s'" % host)

    def peers_tracker(self):
        """
        Tries to retrieve as much information as possible about the
        peers from the currently loaded tracker information.

        It's possible that no tracker information exits for the current
        task and for such situations no state change will occur.
        """

        # retrieves both the announce and the announce list structure from
        # the current info dictionary and uses both of them to create the
        # final list containing the various addresses of trackers, then
        # iterates over each of the trackers to retrieve the information
        # about the various peers associated with the torrent file
        announce = self.info.get("announce", None)
        announce_list = self.info.get("announce-list", [[announce]])
        for tracker in announce_list:
            # iterates over the complete set of tracker urls to try to retrieve
            # the various trackers from each of them
            for tracker_url in tracker:
                # retrieves the first element of the tracker structure as the
                # url of it and then verifies that it references an http based
                # tracker (as that's the only one supported)
                is_http = tracker_url.startswith(("http://", "https://"))
                if not is_http: continue

                # runs the get http retrieval call (blocking call) so that it's
                # possible to retrieve the contents for the announce of the tracker
                # this is an asynchronous call and the on tracker callback will be
                # called at the end of the process with the message
                self.owner.http_client.get(
                    tracker_url,
                    params = dict(
                        info_hash = self.info_hash,
                        peer_id = self.owner.peer_id,
                        port = 6881,
                        uploaded = self.uploaded,
                        downloaded = self.downloaded,
                        left = self.left(),
                        compact = 1,
                        no_peer_id = 0,
                        event = "started",
                        numwant = 50,
                        key = self.owner.get_id()
                    ),
                    on_result = self.on_tracker
                )

                # prints a debug message about the request for peer that was just
                # performed in order to provide some debugging information
                self.owner.debug("Requested peers using '%s'" % tracker_url)

    def peers_file(self):
        for path in PEER_PATHS:
            path = os.path.expanduser(path)
            path = os.path.normpath(path)
            if not os.path.exists(path): continue
            file = open(path, "r")
            for line in file:
                line = line.strip()
                host, port = line.split(":", 1)
                port = int(port)
                peer = dict(ip = host, port = port)
                self.add_peer(peer)

    def connect_peers(self):
        for peer in self.peers: self.connect_peer(peer)

    def disconnect_peers(self):
        connections = copy.copy(self.connections)
        for connection in connections: connection.close(flush = True)

    def connect_peer(self, peer):
        if not peer["new"]: return
        peer["new"] = False
        self.owner.debug("Connecting to peer '%s:%d'" % (peer["ip"], peer["port"]))
        connection = self.owner.client.peer(self, peer["ip"], peer["port"])
        self.connections.append(connection)
        connection.bind("close", self.on_close)
        connection.bind("choked", self.on_choked)
        connection.bind("unchoked", self.on_unchoked)

    def info_string(self):
        return "==== STATUS ====\n" +\
            "peers       := %d\n" % len(self.peers) +\
            "connections := %d\n" % len(self.connections) +\
            "choked      := %d\n" % (len(self.connections) - self.unchoked) +\
            "unchoked    := %d\n" % self.unchoked +\
            "pieces      := %d/%d\n" % (self.stored.marked_pieces, self.stored.total_pieces) +\
            "blocks      := %d/%d\n" % (self.stored.marked_blocks, self.stored.total_blocks) +\
            "pieces miss := %s\n" % self.stored.missing_pieces +\
            "blocks miss := %s\n" % self.stored.missing_blocks +\
            "percent     := %.2f % %\n" % self.percent() +\
            "left        := %d/%d bytes\n" % (self.left(), self.info["length"]) +\
            "speed       := %s/s" % self.speed_s()

    def left(self):
        size = self.info["length"]
        return size - self.downloaded

    def speed(self):
        """
        Retrieves a float number representing the global speed
        of the task in bytes per second, this value is computed
        using the original creation time of the task and so it
        may not represent the most accurate speedup.

        :rtype: float
        :return: The current speed of download, defined as bytes
        per second from the original task creation time.
        """

        current = time.time()
        delta = current - self.start
        bytes_second = self.downloaded / delta
        return bytes_second

    def speed_s(self):
        return netius.common.size_round_unit(
            self.speed(),
            space = True,
            reduce = False
        )

    def percent(self):
        size = self.info["length"]
        return float(self.downloaded) / float(size) * 100.0

    def pop_block(self, bitfield):
        left = self.left()
        is_end = left < THRESHOLD_END
        structure = self.stored if is_end else self.requested
        if not structure: return None
        return structure.pop_block(bitfield, mark = not is_end)

    def push_block(self, index, begin):
        if not self.requested: return
        self.requested.push_block(index, begin)

    def verify_piece(self, index):
        self._verify_piece(index, self.file)

    def confirm_piece(self, index):
        piece_size = self.stored.piece_size(index)
        self.downloaded += piece_size

    def refute_piece(self, index):
        self.requested.mark_piece(index, value = True)
        self.stored.mark_piece(index, value = True)
        self.owner.warning("Refuted piece '%d' (probably invalid)" % index)

    def extend_peers(self, peers):
        for peer in peers: self.add_peer(peer)

    def add_peer(self, peer):
        peer_t = (peer["ip"], peer["port"])
        if peer_t in self.peers_m: return
        peer["time"] = time.time()
        peer["new"] = True
        self.peers_m[peer_t] = peer
        self.peers.append(peer)

    def remove_peer(self, peer):
        peer_t = (peer["ip"], peer["port"])
        if not peer_t in self.peers_m: return
        del self.peers_m[peer_t]
        self.peers.remove(peer)

    def _is_single(self):
        files = self.info.get("files", [])
        return False if files else True

    def _verify_piece(self, index, file):
        piece = self.info["pieces"][index]
        piece_length = self.info["info"]["piece length"]
        file.seek(index * piece_length)
        pending = self.stored.piece_size(index)
        hash = hashlib.sha1()
        while True:
            if pending == 0: break
            count = BLOCK_SIZE if pending > BLOCK_SIZE else pending
            data = file.read(count)
            hash.update(data)
            pending -= count
        digest = hash.digest()
        piece = netius.legacy.bytes(piece)
        if digest == piece: return
        raise netius.DataError("Verifying piece index '%d'" % index)

class TorrentServer(netius.ContainerServer):

    def __init__(self, *args, **kwargs):
        netius.ContainerServer.__init__(self, *args, **kwargs)
        self.peer_id = self._generate_id()
        self.client = netius.clients.TorrentClient(
            thread = False,
            *args,
            **kwargs
        )
        self.http_client = netius.clients.HTTPClient(
            thread = False,
            *args,
            **kwargs
        )
        self.dht_client = netius.clients.DHTClient(
            thread = False,
            *args,
            **kwargs
        )
        self.tasks = []
        self.add_base(self.client)
        self.add_base(self.http_client)
        self.add_base(self.dht_client)

    def cleanup(self):
        netius.ContainerServer.cleanup(self)
        self.cleanup_tasks()
        self.client.destroy()

    def ticks(self):
        netius.ContainerServer.ticks(self)
        for task in self.tasks: task.ticks()

    def download(self, target_path, torrent_path = None, info_hash = None, close = False):
        """
        Starts the "downloading" process of a torrent associated file
        using the defined peer to peer torrent strategy using either
        the provided torrent path as reference or just the info hash
        of the file that is going to be downloaded.

        Note that if only the info hash is provided a DHT based strategy
        is going to be used to retrieve the peers list.

        The returned value is the task entity representing the newly created
        task for the downloading of the requested file, this object may be
        used for the operations and listening of events.

        :type target_path: String
        :param target_path: The path to the directory that will be used to store
        the binary information resulting from the download, this directory may also
        be used to store some temporary information on state of download.
        :type torrent_path: String
        :param torrent_path: The path to the file that contains the torrent
        information that is going to be used for file processing.
        :type info_hash: String
        :param info_hash: The info hash value of the file that is going
        to be downloaded, may be used for magnet torrents (DHT).
        :type close: bool
        :param close: If the server infra-structure should be close (process ends)
        at the end of the download, this is not the default behavior (multiple download).
        :rtype: TorrentTask
        :return: The torrent task object that represents the task that has been
        created for downloading of the requested file.
        """

        def on_complete(task):
            owner = task.owner
            self.remove_task(task)
            if close: owner.close()

        task = TorrentTask(
            self,
            target_path,
            torrent_path = torrent_path,
            info_hash = info_hash
        )
        task.load()
        task.connect_peers()
        task.bind("complete", on_complete)
        self.tasks.append(task)
        return task

    def add_task(self, task):
        self.tasks.append(task)

    def remove_task(self, task):
        task.unload()
        self.tasks.remove(task)

    def cleanup_tasks(self):
        tasks = copy.copy(self.tasks)
        for task in tasks: self.remove_task(task)

    def _generate_id(self):
        random = str(uuid.uuid4())
        random = netius.legacy.bytes(random)
        hash = hashlib.sha1(random)
        digest = hash.hexdigest()
        id = "-%s-%s" % (ID_STRING, digest[:12])
        return id

if __name__ == "__main__":
    import logging

    if len(sys.argv) > 1: file_path = sys.argv[1]
    else: file_path = "\\file.torrent"

    def on_start(server):
        task = server.download("~/Downloads", file_path, close = True)
        task.bind("piece", on_piece)
        task.bind("complete", on_complete)

    def on_piece(task, index):
        percent = task.percent()
        speed_s = task.speed_s()
        left = task.left()
        percent = int(percent)
        print(task.info_string())
        print("[%d%%] - %d bytes (%s/s)" % (percent, left, speed_s))

    def on_complete(task):
        print("Download completed")

    server = TorrentServer(level = logging.DEBUG)
    server.bind("start", on_start)
    server.serve(env = True)
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os

import netius

from . import http2

BOUNDARY = "mjpegboundary"
""" The defualt boundary string value to be used in
case no boundary is provided to the app """

class MJPGServer(http2.HTTP2Server):
    """
    Server class for the creation of an HTTP server for
    the providing of a motion JPEG stream (as in spec).

    This class should only be seen as a foundation and
    proper implementation should be made from this.
    """

    def __init__(self, boundary = BOUNDARY, *args, **kwargs):
        http2.HTTP2Server.__init__(self, *args, **kwargs)
        self.boundary = boundary

    def on_data_http(self, connection, parser):
        http2.HTTP2Server.on_data_http(self, connection, parser)

        headers = [
            ("Content-type", "multipart/x-mixed-replace; boundary=%s" % self.boundary),
            ("Cache-Control", "no-cache"),
            ("Connection", "close"),
            ("Pragma", "no-cache")
        ]

        version_s = parser.version_s
        headers = dict(headers)

        connection.send_header(
            headers = headers,
            version = version_s,
            code = 200,
            code_s = "OK"
        )

        def send(connection):
            self.on_send_mjpg(connection)

            delay = self.get_delay(connection)
            data = self.get_image(connection)

            if not data: self.warning("No image retrieved from provider")
            if not data: data = b""

            data_l = len(data)

            buffer = []
            buffer.append(netius.legacy.bytes("--%s\r\n" % self.boundary))
            buffer.append(b"Content-Type: image/jpeg\r\n")
            buffer.append(netius.legacy.bytes("Content-Length: %d\r\n" % data_l))
            buffer.append(b"\r\n")
            buffer.append(data)
            buffer.append(b"\r\n")

            buffer_d = b"".join(buffer)

            def next(connection):
                def callable(): send(connection)
                self.delay(callable, delay)

            connection.send_part(buffer_d, final = False, callback = next)

        send(connection)

    def on_send_mjpg(self, connection):
        pass

    def get_delay(self, connection):
        return 1

    def get_image(self, connection):
        has_index = hasattr(connection, "index")
        if not has_index: connection.index = 0
        target = connection.index % 2
        connection.index += 1

        base_path = os.path.dirname(__file__)
        extras_path = os.path.join(base_path, "extras")
        file_path = os.path.join(extras_path, "boy_%d.jpg" % target)

        file = open(file_path, "rb")
        try: data = file.read()
        finally: file.close()

        return data

if __name__ == "__main__":
    server = MJPGServer()
    server.serve(env = True)
else:
    __path__ = []
